# -*- coding: utf-8 -*-

"""Tests for the command line interface."""

import json
import logging
import os
import traceback
import unittest

from click.testing import CliRunner

from pybel import Manager, cli
from pybel.constants import METADATA_NAME, PYBEL_CONTEXT_TAG
from pybel.io import from_bel_script, from_nodelink, from_pickle
from pybel.manager.database_io import from_database
from pybel.testing.cases import FleetingTemporaryCacheMixin
from pybel.testing.constants import test_bel_simple, test_bel_thorough
from pybel.testing.mocks import mock_bel_resources
from tests.constants import BelReconstitutionMixin, expected_test_thorough_metadata

log = logging.getLogger(__name__)


@unittest.skip
class TestCli(FleetingTemporaryCacheMixin, BelReconstitutionMixin):
    def setUp(self):
        super(TestCli, self).setUp()
        self.runner = CliRunner()

    @mock_bel_resources
    def test_convert(self, mock_get):
        """Test conversion via the CLI."""
        with self.runner.isolated_filesystem():
            test_csv = os.path.abspath("test.csv")
            test_gpickle = os.path.abspath("test.gpickle")
            test_canon = os.path.abspath("test.bel")

            args = [
                "convert",
                # Input
                "--path",
                test_bel_thorough,
                "--connection",
                self.connection,
                # Outputs
                "--csv",
                test_csv,
                "--pickle",
                test_gpickle,
                "--bel",
                test_canon,
                "--store",
                "--allow-nested",
            ]

            result = self.runner.invoke(cli.main, args)
            self.assertEqual(
                0,
                result.exit_code,
                msg="{}\n{}\n{}".format(
                    result.exc_info[0],
                    result.exc_info[1],
                    traceback.format_tb(result.exc_info[2]),
                ),
            )

            self.assertTrue(os.path.exists(test_csv))

            self.bel_thorough_reconstituted(from_pickle(test_gpickle))
            self.bel_thorough_reconstituted(from_bel_script(test_canon))

            manager = Manager(connection=self.connection)
            self.bel_thorough_reconstituted(
                from_database(expected_test_thorough_metadata[METADATA_NAME], manager=manager)
            )

    @mock_bel_resources
    def test_convert_json(self, mock_get):
        with self.runner.isolated_filesystem():
            test_json = os.path.abspath("test.json")

            args = [
                "convert",
                "--path",
                test_bel_thorough,
                "--json",
                test_json,
                "--connection",
                self.connection,
                "--allow-nested",
            ]

            result = self.runner.invoke(cli.main, args)
            self.assertEqual(0, result.exit_code, msg=result.exc_info)

            with open(test_json) as f:
                self.bel_thorough_reconstituted(from_nodelink(json.load(f)))

    @unittest.skipUnless("NEO_PATH" in os.environ, "Need environmental variable $NEO_PATH")
    @mock_bel_resources
    def test_neo4j_remote(self, mock_get):
        from py2neo import Graph
        from py2neo.database.status import GraphError

        test_context = "PYBEL_TEST_CTX"
        neo_path = os.environ["NEO_PATH"]

        try:
            neo = Graph(neo_path)
            neo.data('match (n)-[r]->() where r.{}="{}" detach delete n'.format(PYBEL_CONTEXT_TAG, test_context))
        except GraphError:
            self.skipTest("Can't query Neo4J ")
        except Exception:
            self.skipTest("Can't connect to Neo4J server")
        else:
            with self.runner.isolated_filesystem():
                args = [
                    "convert",
                    "--path",
                    test_bel_simple,
                    "--connection",
                    self.connection,
                    "--neo",
                    neo_path,
                    "--neo-context",
                    test_context,
                ]
                self.runner.invoke(cli.main, args)

                q = 'match (n)-[r]->() where r.{}="{}" return count(n) as count'.format(PYBEL_CONTEXT_TAG, test_context)
                count = neo.data(q)[0]["count"]
                self.assertEqual(14, count)



# -*- coding: utf-8 -*-

"""Export functions for Machine Learning.

While BEL is a fantastic medium for storing metadata and high granularity
information on edges, machine learning algorithms can not consume BEL graphs
directly. This module provides functions that make inferences and interpretations
of BEL graphs in order to interface with machine learning platforms. One
example where we've done this is `BioKEEN <https://doi.org/10.1093/bioinformatics/btz117>`_,
which uses this module to convert BEL graphs into a format for knowledge graph embeddings.
"""

from .api import to_edgelist, to_triples, to_triples_file  # noqa: F401

__all__ = [
    "to_triples_file",
    "to_triples",
    "to_edgelist",
]



# -*- coding: utf-8 -*-

"""Tests for canonicalization functions."""

import unittest
from typing import Iterable

from pybel import BELGraph
from pybel.canonicalize import _to_bel_lines_body, postpend_location
from pybel.constants import CITATION_TYPE_PUBMED, EXTRACELLULAR, INTRACELLULAR, MODIFIER
from pybel.dsl import (
    Abundance,
    BiologicalProcess,
    ComplexAbundance,
    CompositeAbundance,
    EnumeratedFusionRange,
    Fragment,
    Gene,
    GeneFusion,
    GeneModification,
    Hgvs,
    MicroRna,
    NamedComplexAbundance,
    Pathology,
    Protein,
    ProteinModification,
    ProteinSubstitution,
    Reaction,
    Rna,
    RnaFusion,
    activity,
    degradation,
    secretion,
    translocation,
)
from pybel.language import Entity
from pybel.testing.utils import n
from pybel.utils import canonicalize_edge


class TestCanonicalize(unittest.TestCase):
    def test_postpend_location_failure(self):
        with self.assertRaises(ValueError):
            postpend_location("", dict(name="failure"))

    def test_canonicalize_variant_dsl(self):
        """Use the __str__ functions in the DSL to create BEL instead of external pybel.canonicalize."""
        self.assertEqual('var("p.Val600Glu")', str(Hgvs("p.Val600Glu")))
        self.assertEqual('var("p.Val600Glu")', str(ProteinSubstitution("Val", 600, "Glu")))

        self.assertEqual(
            'pmod(go:0006468 ! "protein phosphorylation")',
            str(ProteinModification("Ph")),
        )
        self.assertEqual("pmod(TEST:Ph)", str(ProteinModification("Ph", namespace="TEST")))
        self.assertEqual(
            "pmod(TEST:Ph, Ser)",
            str(ProteinModification("Ph", namespace="TEST", code="Ser")),
        )
        self.assertEqual(
            "pmod(TEST:Ph, Ser, 5)",
            str(ProteinModification("Ph", namespace="TEST", code="Ser", position=5)),
        )
        self.assertEqual(
            'pmod(GO:"protein phosphorylation", Thr, 308)',
            str(
                ProteinModification(
                    name="protein phosphorylation",
                    namespace="GO",
                    code="Thr",
                    position=308,
                )
            ),
        )

        self.assertEqual('frag("?")', str(Fragment()))
        self.assertEqual('frag("672_713")', str(Fragment(start=672, stop=713)))
        self.assertEqual('frag("?", "descr")', str(Fragment(description="descr")))
        self.assertEqual(
            'frag("672_713", "descr")',
            str(Fragment(start=672, stop=713, description="descr")),
        )

        self.assertEqual('gmod(go:0006306 ! "DNA methylation")', str(GeneModification("Me")))
        self.assertEqual("gmod(TEST:Me)", str(GeneModification("Me", namespace="TEST")))
        self.assertEqual(
            'gmod(GO:"DNA Methylation")',
            str(GeneModification("DNA Methylation", namespace="GO")),
        )

    def test_canonicalize_fusion_range_dsl(self):
        """Test canonicalization of enumerated fusion ranges."""
        self.assertEqual("p.1_15", str(EnumeratedFusionRange("p", 1, 15)))
        self.assertEqual("p.*_15", str(EnumeratedFusionRange("p", "*", 15)))

    def test_Abundance(self):
        """Test canonicalization of abundances."""
        short = Abundance(namespace="CHEBI", name="water")
        self.assertEqual("a(CHEBI:water)", str(short))

        long = Abundance(namespace="CHEBI", name="test name")
        self.assertEqual('a(CHEBI:"test name")', str(long))

    def test_protein_reference(self):
        self.assertEqual("p(HGNC:AKT1)", str(Protein(namespace="HGNC", name="AKT1")))

    def test_gene_reference(self):
        node = Gene(namespace="EGID", name="780")
        self.assertEqual("g(EGID:780)", str(node))

    def test_protein_pmod(self):
        node = Protein(
            name="PLCG1",
            namespace="HGNC",
            variants=[ProteinModification(name="Ph", code="Tyr")],
        )
        self.assertEqual(
            'p(HGNC:PLCG1, pmod(go:0006468 ! "protein phosphorylation", Tyr))',
            str(node),
        )

    def test_protein_fragment(self):
        node = Protein(name="APP", namespace="HGNC", variants=[Fragment(start=672, stop=713)])

        self.assertEqual('p(HGNC:APP, frag("672_713"))', str(node))

    def test_mirna_reference(self):
        self.assertEqual("m(HGNC:MIR1)", str(MicroRna(namespace="HGNC", name="MIR1")))

    def test_rna_fusion_specified(self):
        node = RnaFusion(
            partner_5p=Rna(namespace="HGNC", name="TMPRSS2"),
            range_5p=EnumeratedFusionRange("r", 1, 79),
            partner_3p=Rna(namespace="HGNC", name="ERG"),
            range_3p=EnumeratedFusionRange("r", 312, 5034),
        )
        self.assertEqual('r(fus(HGNC:TMPRSS2, "r.1_79", HGNC:ERG, "r.312_5034"))', str(node))

    def test_rna_fusion_unspecified(self):
        node = RnaFusion(
            partner_5p=Rna(namespace="HGNC", name="TMPRSS2"),
            partner_3p=Rna(namespace="HGNC", name="ERG"),
        )
        self.assertEqual('r(fus(HGNC:TMPRSS2, "?", HGNC:ERG, "?"))', str(node))

    def test_gene_fusion_specified(self):
        node = GeneFusion(
            partner_5p=Gene(namespace="HGNC", name="TMPRSS2"),
            range_5p=EnumeratedFusionRange("c", 1, 79),
            partner_3p=Gene(namespace="HGNC", name="ERG"),
            range_3p=EnumeratedFusionRange("c", 312, 5034),
        )

        self.assertEqual('g(fus(HGNC:TMPRSS2, "c.1_79", HGNC:ERG, "c.312_5034"))', str(node))

    def test_pathology(self):
        node = Pathology(namespace="DO", name="Alzheimer disease")
        self.assertEqual('path(DO:"Alzheimer disease")', str(node))

    def test_bioprocess(self):
        node = BiologicalProcess(namespace="GO", name="apoptosis")
        self.assertEqual("bp(GO:apoptosis)", str(node))

    def test_named_complex_abundance(self):
        node = NamedComplexAbundance(namespace="SCOMP", name="Calcineurin Complex")

        self.assertEqual('complex(SCOMP:"Calcineurin Complex")', str(node))

    def test_complex_abundance(self):
        node = ComplexAbundance(
            members=[
                Protein(namespace="HGNC", name="FOS"),
                Protein(namespace="HGNC", name="JUN"),
            ]
        )
        self.assertEqual("complex(p(HGNC:FOS), p(HGNC:JUN))", str(node))

    def test_composite_abundance(self):
        node = CompositeAbundance(
            members=[
                Protein(namespace="HGNC", name="FOS"),
                Protein(namespace="HGNC", name="JUN"),
            ]
        )
        self.assertEqual("composite(p(HGNC:FOS), p(HGNC:JUN))", str(node))

    def test_reaction(self):
        node = Reaction(
            reactants=[Abundance(namespace="CHEBI", name="A")],
            products=[Abundance(namespace="CHEBI", name="B")],
        )
        self.assertEqual("rxn(reactants(a(CHEBI:A)), products(a(CHEBI:B)))", str(node))


class TestCanonicalizeEdge(unittest.TestCase):
    """This class houses all testing for the canonicalization of edges such that the relation/modifications can be used
    as a second level hash"""

    def setUp(self):
        self.g = BELGraph()
        self.g.annotation_pattern["Species"] = r"\d+"
        self.u = Protein(name="u", namespace="TEST")
        self.v = Protein(name="v", namespace="TEST")
        self.g.add_node_from_data(self.u)
        self.g.add_node_from_data(self.v)

    def get_data(self, k):
        return self.g[self.u][self.v][k]

    def add_edge(self, source_modifier=None, target_modifier=None, annotations=None):
        key = self.g.add_increases(
            self.u,
            self.v,
            evidence=n(),
            citation=n(),
            source_modifier=source_modifier,
            target_modifier=target_modifier,
            annotations=annotations,
        )

        return canonicalize_edge(self.get_data(key))

    def test_failure(self):
        with self.assertRaises(ValueError):
            self.add_edge(source_modifier={MODIFIER: "nope"})

    def test_canonicalize_edge_info(self):
        c1 = self.add_edge(annotations={"Species": "9606"})

        c2 = self.add_edge(annotations={"Species": "9606"})

        c3 = self.add_edge(
            source_modifier=activity("tport"),
        )

        c4 = self.add_edge(
            source_modifier=activity(namespace="go", name="transporter activity", identifier="0005215"),
        )

        self.assertEqual(c1, c2)
        self.assertNotEqual(c1, c3)
        self.assertEqual(c3, c4)

    def test_subject_degradation_location(self):
        self.assertEqual(
            self.add_edge(source_modifier=degradation()),
            self.add_edge(source_modifier=degradation()),
        )

        self.assertEqual(
            self.add_edge(source_modifier=degradation(location=Entity(name="somewhere", namespace="GO"))),
            self.add_edge(source_modifier=degradation(location=Entity(name="somewhere", namespace="GO"))),
        )

        self.assertNotEqual(
            self.add_edge(source_modifier=degradation()),
            self.add_edge(source_modifier=degradation(location=Entity(name="somewhere", namespace="GO"))),
        )

    def test_translocation(self):
        self.assertEqual(
            self.add_edge(source_modifier=secretion()),
            self.add_edge(source_modifier=secretion()),
        )

        self.assertEqual(
            self.add_edge(source_modifier=secretion()),
            self.add_edge(source_modifier=translocation(INTRACELLULAR, EXTRACELLULAR)),
        )


class TestSerializeBEL(unittest.TestCase):
    def setUp(self):
        self.citation = n()
        self.evidence = n()
        self.url = n()
        self.graph = BELGraph()
        self.graph.namespace_url["HGNC"] = self.url

    def _help_check_lines(self, lines: Iterable[str]):
        """Check the given lines match the graph built during the tests."""
        self.assertEqual(lines, list(_to_bel_lines_body(self.graph)))

    def test_simple(self):
        """Test a scenario with a qualified edge, but no annotations."""
        self.graph.add_increases(
            Protein(namespace="HGNC", name="YFG1"),
            Protein(namespace="HGNC", name="YFG"),
            citation=self.citation,
            evidence=self.evidence,
        )

        self.assertEqual(2, self.graph.number_of_nodes())
        self.assertEqual(1, self.graph.number_of_edges())

        expected_lines = [
            f'SET Citation = {{"{CITATION_TYPE_PUBMED}", "{self.citation}"}}\n',
            'SET SupportingText = "{}"'.format(self.evidence),
            "p(HGNC:YFG1) increases p(HGNC:YFG)",
            "UNSET SupportingText",
            "UNSET Citation\n",
            "#" * 80,
        ]

        self._help_check_lines(expected_lines)

    def test_different_key_and_namespace(self):
        key, namespace, value = map(lambda _: n(), range(3))

        self.graph.annotation_curie.add(key)
        self.graph.add_increases(
            Protein(namespace="HGNC", name="YFG1"),
            Protein(namespace="HGNC", name="YFG"),
            citation=self.citation,
            evidence=self.evidence,
            annotations={
                key: Entity(namespace=namespace, identifier=value),
            },
        )

        self.assertEqual(2, self.graph.number_of_nodes())
        self.assertEqual(1, self.graph.number_of_edges())

        expected_lines = [
            f'SET Citation = {{"{CITATION_TYPE_PUBMED}", "{self.citation}"}}\n',
            f'SET SupportingText = "{self.evidence}"',
            f'SET {key} = "{namespace}:{value}"',
            "p(HGNC:YFG1) increases p(HGNC:YFG)",
            f"UNSET {key}",
            "UNSET SupportingText",
            "UNSET Citation\n",
            ("#" * 80),
        ]

        self._help_check_lines(expected_lines)

    def test_single_annotation(self):
        """Test a scenario with a qualified edge, but no annotations."""
        a1, v1 = map(lambda _: n(), range(2))
        self.graph.annotation_list[a1] = {v1}
        self.graph.add_increases(
            Protein(namespace="HGNC", name="YFG1"),
            Protein(namespace="HGNC", name="YFG"),
            citation=self.citation,
            evidence=self.evidence,
            annotations={
                a1: {v1},
            },
        )

        self.assertEqual(2, self.graph.number_of_nodes())
        self.assertEqual(1, self.graph.number_of_edges())

        # Means that only the identifier needs to be written out
        self.assertNotIn(a1, self.graph.annotation_curie)

        expected_lines = [
            f'SET Citation = {{"{CITATION_TYPE_PUBMED}", "{self.citation}"}}\n',
            f'SET SupportingText = "{self.evidence}"',
            f'SET {a1} = "{v1}"',
            "p(HGNC:YFG1) increases p(HGNC:YFG)",
            f"UNSET {a1}",
            "UNSET SupportingText",
            "UNSET Citation\n",
            "#" * 80,
        ]

        self._help_check_lines(expected_lines)

    def test_multiple_annotations(self):
        a1, v1, v2 = map(lambda _: n(), range(3))
        v1, v2 = sorted([v1, v2])

        self.graph.annotation_list[a1] = {v1, v2}
        self.graph.add_increases(
            Protein(namespace="HGNC", name="YFG1"),
            Protein(namespace="HGNC", name="YFG"),
            citation=self.citation,
            evidence=self.evidence,
            annotations={
                a1: {v1, v2},
            },
        )

        self.assertEqual(2, self.graph.number_of_nodes())
        self.assertEqual(1, self.graph.number_of_edges())

        expected_lines = [
            f'SET Citation = {{"{CITATION_TYPE_PUBMED}", "{self.citation}"}}\n',
            f'SET SupportingText = "{self.evidence}"',
            f'SET {a1} = {{"{v1}", "{v2}"}}',
            "p(HGNC:YFG1) increases p(HGNC:YFG)",
            f"UNSET {a1}",
            "UNSET SupportingText",
            "UNSET Citation\n",
            ("#" * 80),
        ]

        self._help_check_lines(expected_lines)



# -*- coding: utf-8 -*-

"""Functions for deleting proteins and genes that are leaves."""

from typing import Iterable

from ...filters.node_filters import get_nodes_by_function
from ...pipeline.decorators import in_place_transformation
from ....constants import GENE, RELATION, RNA, TRANSCRIBED_TO, TRANSLATED_TO
from ....dsl import BaseEntity

__all__ = [
    "prune_protein_rna_origins",
]


def get_gene_leaves(graph) -> Iterable[BaseEntity]:
    """Iterate over all genes who have only one connection, that's a transcription to its RNA.

    :param pybel.BELGraph graph: A BEL graph
    """
    yield from _iterate_leaves(graph, GENE, TRANSCRIBED_TO)


def get_rna_leaves(graph) -> Iterable[BaseEntity]:
    """Iterate over all RNAs who have only one connection, that's a translation to its protein.

    :param pybel.BELGraph graph: A BEL graph
    """
    yield from _iterate_leaves(graph, RNA, TRANSLATED_TO)


def _iterate_leaves(graph, func, relation):
    for node in get_nodes_by_function(graph, func):
        if graph.in_degree(node) != 0:
            continue

        if graph.out_degree(node) != 1:
            continue

        _, _, d = list(graph.out_edges(node, data=True))[0]

        if d[RELATION] == relation:
            yield node


@in_place_transformation
def prune_rna_origins(graph):
    """Delete gene nodes that are only connected to one node, their correspond RNA, by a transcription edge.

    :param pybel.BELGraph graph: A BEL graph
    """
    gene_leaves = list(get_gene_leaves(graph))
    graph.remove_nodes_from(gene_leaves)


@in_place_transformation
def prune_protein_origins(graph):
    """Delete RNA nodes that are only connected to one node - their correspond protein - by a translation edge.

    :param pybel.BELGraph graph: A BEL graph
    """
    rna_leaves = list(get_rna_leaves(graph))
    graph.remove_nodes_from(rna_leaves)


@in_place_transformation
def prune_protein_rna_origins(graph):
    """Delete genes that are only connected to one node, their correspond RNA, by a translation edge.

    :param pybel.BELGraph graph: A BEL graph
    """
    prune_rna_origins(graph)
    prune_protein_origins(graph)



# -*- coding: utf-8 -*-

"""Utilities for Jinja2 templating."""

import os

__all__ = [
    "build_template_environment",
    "build_template_renderer",
]


def build_template_environment(here: str):
    """Build a custom templating environment so Flask apps can get data from lots of different places.

    :param here: Give this the result of :code:`os.path.dirname(os.path.abspath(__file__))`
    :rtype: jinja2.Environment
    """
    from jinja2 import Environment, FileSystemLoader

    loader = FileSystemLoader(os.path.join(here, "templates"))
    environment = Environment(
        autoescape=True,
        loader=loader,
        trim_blocks=False,
    )
    environment.globals["STATIC_PREFIX"] = here + "/static/"
    return environment


def build_template_renderer(path: str):
    """Build a render template function.

    :param path: The location of the current file. Pass it :code:`__file__` like in the example below.

    >>> render_template = build_template_renderer(__file__)
    """
    here = os.path.dirname(os.path.abspath(path))
    template_environment = build_template_environment(here)

    def render_template_enclosure(template_filename: str, **context) -> str:
        """Render a template as a unicode string.

        :param template_filename: The name of the file to render in the template directory
        :param dict context: The variables to template
        """
        return template_environment.get_template(template_filename).render(context)

    render_template_enclosure.environment = template_environment
    return render_template_enclosure



# -*- coding: utf-8 -*-

"""Utilities for node filters."""

from ...constants import MODIFIER
from ...typing import EdgeData

__all__ = [
    "part_has_modifier",
]


def part_has_modifier(edge_data: EdgeData, part: str, modifier: str) -> bool:
    """Return true if the modifier is in the given subject/object part.

    :param edge_data: PyBEL edge data dictionary
    :param part: either :data:`pybel.constants.SUBJECT` or :data:`pybel.constants.OBJECT`
    :param modifier: The modifier to look for
    """
    part_data = edge_data.get(part)

    if part_data is None:
        return False

    return part_data.get(MODIFIER) == modifier



# -*- coding: utf-8 -*-

"""Functions for collapsing proteins, RNAs, microRNAs, and variants to their correspongind genes."""

from collections import defaultdict
from typing import Dict, Set

from .collapse import collapse_nodes
from ..inference import enrich_protein_and_rna_origins
from ...pipeline.decorators import in_place_transformation
from ....constants import RELATION, TRANSCRIBED_TO, TRANSLATED_TO
from ....dsl import BaseEntity

__all__ = [
    "collapse_to_genes",
]


def _build_collapse_to_gene_dict(graph) -> Dict[BaseEntity, Set[BaseEntity]]:
    """Build a collapse dictionary.

    :param pybel.BELGraph graph: A BEL graph
    :return: A dictionary of {node: set of PyBEL node tuples}
    """
    collapse_dict = defaultdict(set)
    r2g = {}

    for gene_node, rna_node, d in graph.edges(data=True):
        if d[RELATION] != TRANSCRIBED_TO:
            continue

        collapse_dict[gene_node].add(rna_node)
        r2g[rna_node] = gene_node

    for rna_node, protein_node, d in graph.edges(data=True):
        if d[RELATION] != TRANSLATED_TO:
            continue

        if rna_node not in r2g:
            raise ValueError("Should complete origin before running this function")

        collapse_dict[r2g[rna_node]].add(protein_node)

    return collapse_dict


@in_place_transformation
def collapse_to_genes(graph):
    """Collapse all protein, RNA, and miRNA nodes to their corresponding gene nodes.

    :param pybel.BELGraph graph: A BEL graph
    """
    enrich_protein_and_rna_origins(graph)
    collapse_dict = _build_collapse_to_gene_dict(graph)
    collapse_nodes(graph, collapse_dict)



# -*- coding: utf-8 -*-

"""Tests for graph operations."""

import unittest

from pybel import BELGraph
from pybel.dsl import protein
from pybel.struct.operations import (
    left_full_join,
    left_node_intersection_join,
    left_outer_join,
    node_intersection,
    union,
)
from pybel.testing.utils import n

p1, p2, p3, p4, p5, p6, p7, p8 = (protein(namespace="HGNC", name=n()) for _ in range(8))


class TestLeftFullJoin(unittest.TestCase):
    """Tests the variants of the left full join, including the exhaustive vs. hash algorithms and calling by function
    or magic functions"""

    def setUp(self):
        """Set up tests for the left full join with two example graphs."""
        g = BELGraph()
        g.add_increases(p1, p2, citation="PMID1", evidence="Evidence 1")

        h = BELGraph()
        h.add_increases(p1, p2, citation="PMID1", evidence="Evidence 1")
        h.add_increases(p1, p2, citation="PMID2", evidence="Evidence 2")
        h.add_increases(p1, p3, citation="PMID1", evidence="Evidence 3")

        self.g = g
        self.h = h

        self._help_check_initial_g(self.g)
        self._help_check_initial_h(self.h)

    def _help_check_initial_g(self, graph: BELGraph):
        """Test the initial G graph."""
        self.assertEqual(2, graph.number_of_nodes(), msg="initial graph G had wrong number of nodes")
        self.assertEqual(1, graph.number_of_edges(), msg="initial graph G had wrong number of edges")

    def _help_check_initial_h(self, graph: BELGraph):
        """Test the initial H graph."""
        self.assertEqual(3, graph.number_of_nodes(), msg="initial graph H had wrong number of nodes")
        self.assertEqual(3, graph.number_of_edges(), msg="initial graph H had wrong number of edges")

    def _help_check_result(self, j: BELGraph):
        """Help check the result of left joining H into G.

        :param j: The resulting graph from G += H
        """
        self.assertEqual(3, j.number_of_nodes())
        self.assertEqual(
            3,
            j.number_of_edges(),
            msg="G edges:\n{}".format("\n".join(map(str, j.edges(data=True)))),
        )

    def test_function(self):
        """Test full joining two networks using the function."""
        left_full_join(self.g, self.h)
        self._help_check_result(self.g)
        self._help_check_initial_h(self.h)

    def test_full_join_with_isolated_nodes(self):
        """Test what happens when there are isolated nodes."""
        a = BELGraph()
        a.add_increases(p1, p2, citation=n(), evidence=n())
        a.add_node_from_data(p4)
        b = BELGraph()
        b.add_increases(p2, p3, citation=n(), evidence=n())
        b.add_node_from_data(p5)
        left_full_join(a, b)
        for node in p1, p2, p3, p4, p5:
            self.assertIn(node, a)

    def test_in_place_operator_failure(self):
        """Test that using the wrong type with the in-place addition operator raises an error."""
        with self.assertRaises(TypeError):
            self.g += None

    def test_in_place_operator(self):
        """Test full joining two networks using the BELGraph in-place addition operator."""
        self.g += self.h
        self._help_check_result(self.g)
        self._help_check_initial_h(self.h)

    def test_operator_failure(self):
        """Test that using the wrong type with the addition operator raises an error."""
        with self.assertRaises(TypeError):
            self.g + None

    def test_operator(self):
        """Test full joining two networks using the BELGraph addition operator."""
        j = self.g + self.h
        self._help_check_result(j)
        self._help_check_initial_g(self.g)
        self._help_check_initial_h(self.h)

    def test_union_failure(self):
        """Test that the union of no graphs raises a value error."""
        with self.assertRaises(ValueError):
            union([])

    def test_union_trivial(self):
        """Test that the union of a single graph returns that graph."""
        res = union([self.g])
        self.assertEqual(self.g, res)

    def test_union(self):
        """Test that the union of a pair of graphs is the same as the full join."""
        j = union([self.g, self.h])
        self._help_check_result(j)
        self._help_check_initial_g(self.g)
        self._help_check_initial_h(self.h)


class TestLeftFullOuterJoin(unittest.TestCase):
    def setUp(self):
        g = BELGraph()

        g.add_edge(p1, p2)

        h = BELGraph()
        h.add_edge(p1, p3)
        h.add_edge(p1, p4)

        h.add_edge(p5, p6)
        h.add_node(p7)

        self.g = g
        self.h = h

    def _help_check_initial_g(self, g):
        self.assertEqual(2, g.number_of_nodes())
        self.assertEqual({p1, p2}, set(g))
        self.assertEqual(1, g.number_of_edges())
        self.assertEqual({(p1, p2)}, set(g.edges()))

    def _help_check_initial_h(self, h):
        self.assertEqual(6, h.number_of_nodes())
        self.assertEqual({p1, p3, p4, p5, p6, p7}, set(h))
        self.assertEqual(3, h.number_of_edges())
        self.assertEqual({(p1, p3), (p1, p4), (p5, p6)}, set(h.edges()))

    def _help_check_result(self, j):
        """After H has been full outer joined into G, this is what it should be"""
        self.assertEqual(4, j.number_of_nodes())
        self.assertEqual({p1, p2, p3, p4}, set(j))
        self.assertEqual(3, j.number_of_edges())
        self.assertEqual({(p1, p2), (p1, p3), (p1, p4)}, set(j.edges()))

    def test_in_place_type_failure(self):
        with self.assertRaises(TypeError):
            self.g &= None

    def test_type_failure(self):
        with self.assertRaises(TypeError):
            self.g & None

    def test_magic(self):
        # left_outer_join(g, h)
        self.g &= self.h
        self._help_check_initial_h(self.h)
        self._help_check_result(self.g)

    def test_operator(self):
        # left_outer_join(g, h)
        j = self.g & self.h
        self._help_check_initial_h(self.h)
        self._help_check_initial_g(self.g)
        self._help_check_result(j)

    def test_left_outer_join(self):
        left_outer_join(self.g, self.h)
        self._help_check_initial_h(self.h)
        self._help_check_result(self.g)

    def test_left_outer_exhaustive_join(self):
        self.g &= self.h
        left_outer_join(self.g, self.h)
        self._help_check_initial_h(self.h)
        self._help_check_result(self.g)


class TestInnerJoin(unittest.TestCase):
    """Tests various graph merging procedures"""

    def setUp(self):
        g = BELGraph()

        g.add_edge(p1, p2)
        g.add_edge(p1, p3)
        g.add_edge(p8, p3)

        h = BELGraph()
        h.add_edge(p1, p3)
        h.add_edge(p1, p4)
        h.add_edge(p5, p6)
        h.add_node(p7)

        self.g = g
        self.h = h

    def _help_check_initialize_g(self, graph):
        self.assertEqual(4, graph.number_of_nodes())
        self.assertEqual(3, graph.number_of_edges())

    def _help_check_initialize_h(self, graph):
        self.assertEqual(6, graph.number_of_nodes())
        self.assertEqual({p1, p3, p4, p5, p6, p7}, set(graph))
        self.assertEqual(3, graph.number_of_edges())
        self.assertEqual({(p1, p3), (p1, p4), (p5, p6)}, set(graph.edges()))

    def test_initialize(self):
        self._help_check_initialize_g(self.g)
        self._help_check_initialize_h(self.h)

    def _help_check_join(self, j):
        self.assertEqual(2, j.number_of_nodes())
        self.assertEqual({p1, p3}, set(j))
        self.assertEqual(1, j.number_of_edges())
        self.assertEqual(
            {
                (p1, p3),
            },
            set(j.edges()),
        )

    def test_in_place_type_failure(self):
        with self.assertRaises(TypeError):
            self.g ^ None

    def test_type_failure(self):
        with self.assertRaises(TypeError):
            self.g ^= None

    def test_magic(self):
        j = self.g ^ self.h
        self._help_check_join(j)
        self._help_check_initialize_h(self.h)
        self._help_check_initialize_g(self.g)

    def test_left_node_intersection_join(self):
        j = left_node_intersection_join(self.g, self.h)
        self._help_check_join(j)
        self._help_check_initialize_h(self.h)
        self._help_check_initialize_g(self.g)

    def test_node_intersection(self):
        j = node_intersection([self.h, self.g])
        self._help_check_join(j)
        self._help_check_initialize_h(self.h)
        self._help_check_initialize_g(self.g)

    def test_intersection_failure(self):
        with self.assertRaises(ValueError):
            node_intersection([])

    def test_intersection_trivial(self):
        res = node_intersection([self.g])
        self.assertEqual(self.g, res)



# -*- coding: utf-8 -*-

"""Conversion functions for BEL graphs with `GraphML <https://en.wikipedia.org/wiki/GraphML>`_."""

from typing import BinaryIO, Optional, Union

import networkx as nx

from ..canonicalize import edge_to_tuple
from ..constants import RELATION
from ..struct import BELGraph

__all__ = [
    "to_graphml",
]


def to_graphml(graph: BELGraph, path: Union[str, BinaryIO], schema: Optional[str] = None) -> None:
    """Write a graph to a GraphML XML file using :func:`networkx.write_graphml`.

    :param graph: BEL Graph
    :param path: Path to the new exported file
    :param schema: Type of export. Currently supported: "simple" and "umbrella".

    The .graphml file extension is suggested so Cytoscape can recognize it.
    By default, this function exports using the PyBEL schema of including modifier information into the edges.
    As an alternative, this function can also distinguish between
    """
    if schema is None or schema == "simple":
        rv = _to_graphml_simple(graph)
    elif schema == "umbrella":
        rv = _to_graphml_umbrella(graph)
    else:
        raise ValueError("Unhandled schema: {}".format(schema))

    nx.write_graphml(rv, path)


def _to_graphml_simple(graph: BELGraph) -> nx.MultiDiGraph:
    """Convert a BEL graph to a simple graph.

    :param graph: A BEL graph
    """
    rv = nx.MultiDiGraph()

    for node in graph:
        rv.add_node(node.as_bel(), function=node.function)

    for u, v, key, edge_data in graph.edges(data=True, keys=True):
        u_key, v_key = u.as_bel(), v.as_bel()
        rv.add_edge(
            u_key,
            v_key,
            key=key,
            interaction=edge_data[RELATION],
            bel=graph.edge_to_bel(u, v, edge_data),
        )

    return rv


def _to_graphml_umbrella(graph: BELGraph) -> nx.MultiDiGraph:
    """Convert a BEL graph to a new graph the nodes as original BEL terms strings.

    :param graph: A BEL graph
    """
    rv = nx.MultiDiGraph()

    for u, v, key, edge_data in graph.edges(data=True, keys=True):
        u_key, _, v_key = edge_to_tuple(u, v, edge_data)

        rv.add_edge(
            u_key,
            v_key,
            key=key,
            relation=edge_data[RELATION],
            bel=graph.edge_to_bel(u, v, edge_data),
        )

    return rv



# -*- coding: utf-8 -*-

"""General-use induction functions."""

import networkx as nx

from ..filters.edge_filters import filter_edges
from ..filters.typing import EdgePredicates
from ..pipeline import (
    in_place_transformation,
    transformation,
    uni_in_place_transformation,
)
from ..utils import update_metadata

__all__ = [
    "remove_isolated_nodes",
    "remove_isolated_nodes_op",
    "expand_by_edge_filter",
]


@in_place_transformation
def remove_isolated_nodes(graph):
    """Remove isolated nodes from the network, in place.

    :param pybel.BELGraph graph: A BEL graph
    """
    nodes = list(nx.isolates(graph))
    graph.remove_nodes_from(nodes)


@transformation
def remove_isolated_nodes_op(graph):
    """Build a new graph excluding the isolated nodes.

    :param pybel.BELGraph graph: A BEL graph
    :rtype: pybel.BELGraph
    """
    rv = graph.copy()
    nodes = list(nx.isolates(rv))
    rv.remove_nodes_from(nodes)
    return rv


@uni_in_place_transformation
def expand_by_edge_filter(source, target, edge_predicates: EdgePredicates):
    """Expand a target graph by edges in the source matching the given predicates.

    :param pybel.BELGraph source: A BEL graph
    :param pybel.BELGraph target: A BEL graph
    :param edge_predicates: An edge predicate or list of edge predicates
    :return: A BEL sub-graph induced over the edges passing the given filters
    :rtype: pybel.BELGraph
    """
    target.add_edges_from(
        (u, v, k, source[u][v][k]) for u, v, k in filter_edges(source, edge_predicates=edge_predicates)
    )

    update_metadata(source, target)
    # TODO smarter ways of ensuring metadata



# -*- coding: utf-8 -*-

"""Classes for DSL nodes."""

import hashlib
from abc import ABCMeta, abstractmethod
from operator import methodcaller
from typing import Iterable, List, Optional, Set, Union

from .exc import (
    InferCentralDogmaException,
    ListAbundanceEmptyException,
    ReactionEmptyException,
)
from ..constants import (
    ABUNDANCE,
    BIOPROCESS,
    COMPLEX,
    COMPOSITE,
    CONCEPT,
    FRAGMENT,
    FRAGMENT_DESCRIPTION,
    FRAGMENT_MISSING,
    FRAGMENT_START,
    FRAGMENT_STOP,
    FUNCTION,
    FUSION,
    FUSION_MISSING,
    FUSION_REFERENCE,
    FUSION_START,
    FUSION_STOP,
    GENE,
    GMOD,
    HGVS,
    KIND,
    MEMBERS,
    MIRNA,
    PARTNER_3P,
    PARTNER_5P,
    PATHOLOGY,
    PMOD,
    PMOD_CODE,
    PMOD_ORDER,
    PMOD_POSITION,
    POPULATION,
    PRODUCTS,
    PROTEIN,
    RANGE_3P,
    RANGE_5P,
    REACTANTS,
    REACTION,
    RNA,
    VARIANTS,
    XREFS,
    rev_abundance_labels,
)
from ..language import Entity, gmod_mappings, pmod_mappings

__all__ = [
    # Base Classes
    "Entity",
    "BaseEntity",
    "BaseAbundance",
    "ListAbundance",
    # Named entities
    "Abundance",
    "BiologicalProcess",
    "Pathology",
    "Population",
    "NamedComplexAbundance",
    # Central Dogma Stuff
    "CentralDogma",
    "Gene",
    "Transcribable",
    "Rna",
    "MicroRna",
    "Protein",
    # Fusions
    "FusionBase",
    "ProteinFusion",
    "RnaFusion",
    "GeneFusion",
    # Fusion Ranges
    "FusionRangeBase",
    "EnumeratedFusionRange",
    "MissingFusionRange",
    # Variants
    "Variant",
    "EntityVariant",
    "ProteinModification",
    "GeneModification",
    "Hgvs",
    "HgvsReference",
    "HgvsUnspecified",
    "ProteinSubstitution",
    "Fragment",
    # List Entities
    "ComplexAbundance",
    "CompositeAbundance",
    "Reaction",
]

# A methodcaller for the key argument of sorted()
_as_bel = methodcaller("as_bel")


class BaseEntity(dict, metaclass=ABCMeta):
    """This is the superclass for all BEL terms.

    A BEL term has three properties:

    1. It has a type. Subclasses of this function should set the class variable
       ``function``.
    2. It can be converted to BEL. Note, this is an abstract class, so all sub-classes
       must implement this functionality in ``as_bel()``.
    3. It can be hashed, based on the BEL conversion
    """

    function = ...

    def __init__(self) -> None:
        super().__init__(**{FUNCTION: self.function})
        self._md5 = None

    @property
    def _bel_function(self) -> str:
        return rev_abundance_labels[self.function]

    @abstractmethod
    def as_bel(self, use_identifiers: bool = True) -> str:
        """Return this entity as a BEL string."""

    @property
    def md5(self) -> str:
        """Get the MD5 hash of this node."""
        if self._md5 is None:
            self._md5 = hashlib.md5(self.as_bel().encode("utf8")).hexdigest()  # noqa: S303
        return self._md5

    def __hash__(self):  # noqa: D105
        return hash(self.as_bel())

    def __eq__(self, other):
        return isinstance(other, BaseEntity) and self.as_bel() == other.as_bel()

    def __repr__(self):
        return "<BEL {bel}>".format(bel=self.as_bel(use_identifiers=True))

    def __str__(self):  # noqa: D105
        return self.as_bel()

    @property
    def safe_label(self) -> str:
        """Get the safe label for the node (name or BEL)."""
        if isinstance(self, CentralDogma) and self.variants:
            return self.as_bel()
        if isinstance(self, BaseConcept):
            return self.curie
        return self.as_bel()


class BaseConcept(dict):
    """A dictionary containing a concept entry."""

    @property
    def entity(self) -> Entity:  # noqa:D401
        """This node's concept."""
        return self[CONCEPT]

    @property
    def xrefs(self) -> List[Entity]:  # noqa:D401
        """Alternative identifiers for the node's concept."""
        return self.get(XREFS, [])

    @property
    def namespace(self) -> str:  # noqa:D401
        """The namespace of this abundance."""
        return self.entity.namespace

    @property
    def name(self) -> Optional[str]:  # noqa:D401
        """The name of this abundance."""
        return self.entity.name

    @property
    def identifier(self) -> Optional[str]:  # noqa:D401
        """The identifier of this abundance."""
        return self.entity.identifier

    @property
    def curie(self):  # noqa: D401
        """The CURIE-style identifier for this node."""
        return self.entity.curie

    @property
    def obo(self) -> str:  # noqa: D401
        """The OBO-style identifier for this node."""
        return self.entity.obo


class BaseAbundance(BaseEntity, BaseConcept):
    """The superclass for all named BEL terms.

    A named BEL term has:

    1. A type (taken care of by being a subclass of :class:`BaseEntity`)
    2. A named :class:`Entity`. Though this doesn't directly inherit from
       :class:`Entity`, it creates one internally using the namespace,
       identifier, and name. Ideally, both the identifier and name are given.
       If one is missing, it can be looked up with :func:`pybel.grounding.ground`
    3. An optional list of xrefs, corresponding to the whole entity,
       not just the namespace/name. For example, the BEL term
       ``p(HGNC:APP, frag(672_713)`` could xref CHEBI:64647.
    """

    def __init__(
        self,
        namespace: str,
        name: Optional[str] = None,
        identifier: Optional[str] = None,
        xrefs: Optional[List[Entity]] = None,
    ) -> None:
        """Build an abundance from a function, namespace, and a name and/or identifier.

        :param namespace: The name of the namespace
        :param name: The name of this abundance
        :param identifier: The database identifier for this abundance
        :param xrefs: Alternate identifiers for the entity
        """
        super().__init__()
        self[CONCEPT] = Entity(
            namespace=namespace,
            name=name,
            identifier=identifier,
        )
        if xrefs:
            self[XREFS] = xrefs

    def as_bel(self, use_identifiers: bool = True) -> str:
        """Return this node as a BEL string."""
        return "{}({})".format(
            self._bel_function,
            self.obo if use_identifiers and self.entity.identifier and self.entity.name else self.curie,
        )


class Abundance(BaseAbundance):
    """Builds an abundance node.

    >>> from pybel.dsl import Abundance
    >>> Abundance(namespace='CHEBI', name='water')
    """

    function = ABUNDANCE


class BiologicalProcess(BaseAbundance):
    """Builds a biological process node.

    >>> from pybel.dsl import BiologicalProcess
    >>> BiologicalProcess(namespace='GO', name='apoptosis')
    """

    function = BIOPROCESS


class Pathology(BaseAbundance):
    """Build a pathology node.

    >>> from pybel.dsl import Pathology
    >>> Pathology(namespace='DO', name='Alzheimer Disease')
    """

    function = PATHOLOGY


class Population(BaseAbundance):
    """Builds a population node.

    >>> from pybel.dsl import Population
    >>> Population(namespace='uberon', name='blood')
    """

    function = POPULATION


class Variant(dict, metaclass=ABCMeta):
    """The superclass for variant dictionaries."""

    def __init__(self, kind: str) -> None:
        """Build the variant data dictionary.

        :param kind: The kind of variant
        """
        super().__init__({KIND: kind})

    @abstractmethod
    def as_bel(self, use_identifiers: bool = True) -> str:
        """Return this variant as a BEL string."""

    def __str__(self):  # noqa: D105
        return self.as_bel()


class CentralDogma(BaseAbundance):
    """The base class for "central dogma" abundances (i.e., genes, miRNAs, RNAs, and proteins)."""

    def __init__(
        self,
        namespace: str,
        name: Optional[str] = None,
        identifier: Optional[str] = None,
        xrefs: Optional[List[Entity]] = None,
        variants: Union[None, Variant, Iterable[Variant]] = None,
    ) -> None:
        """Build a node for a gene, RNA, miRNA, or protein.

        :param namespace: The name of the database used to identify this entity
        :param name: The database's preferred name or label for this entity
        :param identifier: The database's identifier for this entity
        :param xrefs: Alternative database cross references
        :param variants: An optional variant or list of variants
        """
        super().__init__(namespace=namespace, name=name, identifier=identifier, xrefs=xrefs)

        if isinstance(variants, Variant):
            self[VARIANTS] = [variants]
        elif isinstance(variants, (list, tuple, set)):
            self[VARIANTS] = sorted(variants, key=_as_bel)

    @property
    def variants(self) -> Optional[List[Variant]]:
        """Return this entity's variants, if they exist."""
        return self.get(VARIANTS)

    def as_bel(self, use_identifiers: bool = True) -> str:
        """Return this node as a BEL string."""
        if not self.variants:
            return super().as_bel(use_identifiers=use_identifiers)

        variants_canon = sorted([variant.as_bel(use_identifiers=use_identifiers) for variant in self.variants])

        return "{}({}, {})".format(
            self._bel_function,
            self.obo if use_identifiers and self.entity.identifier and self.entity.name else self.curie,
            ", ".join(variants_canon),
        )

    def get_parent(self) -> Optional["CentralDogma"]:
        """Get the parent, or none if it's already a reference node.

        >>> from pybel.dsl import Protein, Fragment
        >>> ab42 = Protein(name='APP', namespace='HGNC', variants=[Fragment(start=672, stop=713)])
        >>> app = ab42.get_parent()
        >>> assert 'p(HGNC:APP)' == app.as_bel()
        """
        if VARIANTS not in self:
            return None

        return self.__class__(
            namespace=self.namespace,
            name=self.name,
            identifier=self.identifier,
            xrefs=self.xrefs,
        )

    def with_variants(self, variants: Union[Variant, List[Variant]]) -> "CentralDogma":
        """Create a new entity with the given variants.

        :param variants: An optional variant or list of variants

        >>> from pybel.dsl import Protein, Fragment
        >>> app = Protein(name='APP', namespace='HGNC')
        >>> ab42 = app.with_variants([Fragment(start=672, stop=713)])
        >>> assert 'p(HGNC:APP, frag(672_713))' == ab42.as_bel()
        """
        return self.__class__(
            namespace=self.namespace,
            name=self.name,
            identifier=self.identifier,
            xrefs=self.xrefs,
            variants=variants,
        )


class EntityVariant(Variant, BaseConcept):
    """A variant that contains a reference."""

    function = ...

    def __init__(
        self,
        name: str,
        namespace: Optional[str] = None,
        identifier: Optional[str] = None,
        xrefs: Optional[List[Entity]] = None,
    ) -> None:
        """Build a variant that has a reference.

        :param name: The name of the modification
        :param namespace: The namespace to which the name of this modification belongs
        :param identifier: The identifier of the name of the modification
        :param xrefs: Alternative database xrefs

        Either the name or the identifier must be used. If the namespace is omitted, it is assumed that a name is
        specified from the BEL default namespace.
        """
        super().__init__(kind=self.function)

        self[CONCEPT] = Entity(
            namespace=namespace,
            name=name,
            identifier=identifier,
        )
        if xrefs:
            self["xref"] = xrefs


class ProteinModification(EntityVariant):
    """Build a protein modification variant dictionary."""

    function = PMOD

    def __init__(
        self,
        name: str,
        code: Optional[str] = None,
        position: Optional[int] = None,
        namespace: Optional[str] = None,
        identifier: Optional[str] = None,
        xrefs: Optional[List[Entity]] = None,
    ) -> None:
        """Build a protein modification variant data dictionary.

        :param name: The name of the modification
        :param code: The three letter amino acid code for the affected residue. Capital first letter.
        :param position: The position of the affected residue
        :param namespace: The namespace to which the name of this modification belongs
        :param identifier: The identifier of the name of the modification
        :param xrefs: Alternative database xrefs

        Either the name or the identifier must be used. If the namespace is omitted, it is assumed that a name is
        specified from the BEL default namespace.

        Example from BEL default namespace:

        >>> from pybel.dsl import ProteinModification
        >>> ProteinModification('Ph', code='Thr', position=308)

        Example from custom namespace:

        >>> from pybel.dsl import ProteinModification
        >>> ProteinModification(name='protein phosphorylation', namespace='GO', code='Thr', position=308)

        Example from custom namespace additionally qualified with identifier:

        >>> from pybel.dsl import ProteinModification
        >>> ProteinModification(name='protein phosphorylation', namespace='GO',
        >>>                     identifier='0006468', code='Thr', position=308)

        """
        if name and not namespace and not identifier:
            x = pmod_mappings[name]["xrefs"][0]
            namespace, identifier, name = x.namespace, x.identifier, x.name

        super().__init__(
            name=name,
            namespace=namespace,
            identifier=identifier,
            xrefs=xrefs,
        )

        if code:
            self[PMOD_CODE] = code

        if position:
            self[PMOD_POSITION] = position

    def as_bel(self, use_identifiers: bool = True) -> str:
        """Return this protein modification variant as a BEL string."""
        if use_identifiers and self.entity.identifier and self.entity.name:
            x = self.entity.obo
        else:
            x = self.entity.curie

        return "pmod({}{})".format(
            x,
            "".join(", {}".format(self[x]) for x in PMOD_ORDER[2:] if x in self),
        )


class GeneModification(EntityVariant):
    """Build a gene modification variant dictionary."""

    function = GMOD

    def __init__(
        self,
        name: str,
        namespace: Optional[str] = None,
        identifier: Optional[str] = None,
        xrefs: Optional[List[Entity]] = None,
    ) -> None:
        """Build a protein modification variant data dictionary.

        :param name: The name of the modification
        :param namespace: The namespace to which the name of this modification belongs
        :param identifier: The identifier of the name of the modification
        :param xrefs: Alternative database xrefs

        Either the name or the identifier must be used. If the namespace is omitted, it is assumed that a name is
        specified from the BEL default namespace.

        Example from BEL default namespace:

        >>> from pybel.dsl import GeneModification
        >>> GeneModification(name='Me')

        Example from custom namespace:

        >>> from pybel.dsl import GeneModification
        >>> GeneModification(name='DNA methylation', namespace='GO', identifier='0006306')

        """
        if name and not namespace and not identifier:
            x = gmod_mappings[name]["xrefs"][0]
            namespace, identifier, name = x.namespace, x.identifier, x.name

        super().__init__(
            name=name,
            namespace=namespace,
            identifier=identifier,
            xrefs=xrefs,
        )

    def as_bel(self, use_identifiers: bool = True) -> str:
        """Return this gene modification variant as a BEL string."""
        if use_identifiers and self.entity.identifier and self.entity.name:
            x = self.entity.obo
        else:
            x = self.entity.curie

        return "gmod({})".format(x)


class Hgvs(Variant):
    """Builds a HGVS variant dictionary."""

    def __init__(self, variant: str) -> None:
        """Build an HGVS variant data dictionary.

        :param variant: The HGVS variant string

        >>> from pybel.dsl import Protein, Hgvs
        >>> Protein(namespace='HGNC', name='AKT1', variants=[Hgvs('p.Ala127Tyr')])
        """
        super().__init__(kind=HGVS)
        self[HGVS] = variant

    @property
    def variant(self) -> str:  # noqa: D401
        """The HGVS variant string."""
        return self[HGVS]

    def as_bel(self, use_identifiers: bool = True) -> str:
        """Return this HGVS variant as a BEL string."""
        return 'var("{}")'.format(self.variant)


class HgvsReference(Hgvs):
    """Represents the "reference" variant in HGVS."""

    def __init__(self) -> None:
        super().__init__(variant="=")


class HgvsUnspecified(Hgvs):
    """Represents an unspecified variant in HGVS."""

    def __init__(self) -> None:
        super().__init__(variant="?")


class ProteinSubstitution(Hgvs):
    """A protein substitution variant."""

    def __init__(self, from_aa: str, position: int, to_aa: str) -> None:
        """Build an HGVS variant data dictionary for the given protein substitution.

        :param from_aa: The 3-letter amino acid code of the original residue
        :param position: The position of the residue
        :param to_aa: The 3-letter amino acid code of the new residue

        >>> from pybel.dsl import Protein, ProteinSubstitution
        >>> Protein(namespace='HGNC', name='AKT1', variants=[ProteinSubstitution('Ala', 127, 'Tyr')])

        """
        super().__init__("p.{}{}{}".format(from_aa, position, to_aa))


class Fragment(Variant):
    """Represent the information about a protein fragment."""

    def __init__(
        self,
        start: Union[None, int, str] = None,
        stop: Union[None, int, str] = None,
        description: Optional[str] = None,
    ) -> None:
        """Build a protein fragment data dictionary.

        :param start: The starting position
        :param stop: The stopping position
        :param description: An optional description

        Example of specified fragment:

        >>> from pybel.dsl import Protein, Fragment
        >>> Protein(name='APP', namespace='HGNC', variants=[Fragment(start=672, stop=713)])

        Example of unspecified fragment:

        >>> from pybel.dsl import Protein, Fragment
        >>> Protein(name='APP', namespace='HGNC', variants=[Fragment()])

        """
        super().__init__(kind=FRAGMENT)

        if start and stop:
            self[FRAGMENT_START] = start
            self[FRAGMENT_STOP] = stop
        else:
            self[FRAGMENT_MISSING] = "?"

        if description:
            self[FRAGMENT_DESCRIPTION] = description

    @property
    def range(self) -> str:
        """Get the range of this fragment."""
        if FRAGMENT_MISSING in self:
            return "?"

        return "{}_{}".format(self[FRAGMENT_START], self[FRAGMENT_STOP])

    def as_bel(self, use_identifiers=False) -> str:
        """Return this fragment variant as a BEL string."""
        res = '"{}"'.format(self.range)

        if FRAGMENT_DESCRIPTION in self:
            res += ', "{}"'.format(self[FRAGMENT_DESCRIPTION])

        return "frag({})".format(res)


class Gene(CentralDogma):
    """Builds a gene node."""

    function = GENE

    def get_rna(self) -> "Rna":
        """Get the corresponding RNA."""
        if self.variants:
            raise InferCentralDogmaException("can not get gene for variant")
        return Rna(
            namespace=self.namespace,
            name=self.name,
            identifier=self.identifier,
            xrefs=self.xrefs,
        )


class Transcribable(CentralDogma):
    """A base class for RNA and micro-RNA to share getting of their corresponding genes."""

    def get_gene(self) -> Gene:
        """Get the corresponding gene or raise an exception if it's not the reference node.

        :raises: InferCentralDogmaException
        """
        if self.variants:
            raise InferCentralDogmaException("can not get gene for variant")

        return Gene(
            namespace=self.namespace,
            name=self.name,
            identifier=self.identifier,
            xrefs=self.xrefs,
        )


class Rna(Transcribable):
    """Builds an RNA node.

    Example: AKT1 protein coding gene's RNA:

    >>> from pybel.dsl import Rna
    >>> Rna(namespace='HGNC', name='AKT1', identifier='391')

    Non-coding RNAs can also be encoded such as `U85 <https://www-snorna.biotoul.fr/plus.php?id=U85>`_:

    >>> from pybel.dsl import Rna
    >>> Rna(namespace='SNORNABASE', identifier='SR0000073')
    """

    function = RNA


class MicroRna(Transcribable):
    """Represents an micro-RNA.

    Human miRNA's are listed on HUGO's `MicroRNAs (MIR) <https://www.genenames.org/cgi-bin/genefamilies/set/476>`_
    gene family.

    MIR1-1 from `HGNC <https://www.genenames.org/cgi-bin/gene_symbol_report?hgnc_id=31499>`_:

    >>> from pybel.dsl import MicroRna
    >>> MicroRna(namespace='HGNC', name='MIR1-1', identifier='31499')

    MIR1-1 from `miRBase <http://www.mirbase.org/cgi-bin/mirna_entry.pl?acc=MI0000651>`_:

    >>> from pybel.dsl import MicroRna
    >>> MicroRna(namespace='MIRBASE', identifier='MI0000651')

    MIR1-1 from `Entrez Gene <https://view.ncbi.nlm.nih.gov/gene/406904>`_

    >>> from pybel.dsl import MicroRna
    >>> MicroRna(namespace='ENTREZ', identifier='406904')
    """

    function = MIRNA


class Protein(CentralDogma):
    """Builds a protein node.

    Example: AKT

    >>> from pybel.dsl import Protein
    >>> Protein(namespace='HGNC', name='AKT1')

    Example: AKT with optionally included HGNC database identifier

    >>> from pybel.dsl import Protein
    >>> Protein(namespace='HGNC', name='AKT1', identifier='391')

    Example: AKT with phosphorylation

    >>> from pybel.dsl import Protein, ProteinModification
    >>> Protein(namespace='HGNC', name='AKT', variants=[ProteinModification('Ph', code='Thr', position=308)])
    """

    function = PROTEIN

    def get_rna(self) -> Rna:
        """Get the corresponding RNA or raise an exception if it's not the reference node.

        :raises: InferCentralDogmaException
        """
        if self.variants:
            raise InferCentralDogmaException("can not get rna for variant")

        return Rna(
            namespace=self.namespace,
            name=self.name,
            identifier=self.identifier,
            xrefs=self.xrefs,
        )


def _entity_list_as_bel(entities: Iterable[BaseEntity], use_identifiers: bool = True) -> str:
    """Stringify a list of BEL entities."""
    return ", ".join(e.as_bel(use_identifiers=use_identifiers) for e in entities)


def _help_named(self, namespace, identifier, name, xrefs):
    if namespace:
        self[CONCEPT] = Entity(
            namespace=namespace,
            name=name,
            identifier=identifier,
        )
        if xrefs:
            self[XREFS] = xrefs


class Reaction(BaseEntity):
    """Build a reaction node."""

    function = REACTION

    def __init__(
        self,
        reactants: Union[BaseAbundance, Iterable[BaseAbundance]],
        products: Union[BaseAbundance, Iterable[BaseAbundance]],
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        identifier: Optional[str] = None,
        xrefs: Optional[List[Entity]] = None,
    ) -> None:
        """Build a reaction node.

        :param reactants: A list of PyBEL node data dictionaries representing the reactants
        :param products: A list of PyBEL node data dictionaries representing the products
        :param namespace: The namespace from which the name originates
        :param name: The name of the complex
        :param identifier: The identifier in the namespace in which the name originates
        :param xrefs: Alternate identifiers for the entity if it is named

        >>> from pybel.dsl import Reaction, Protein, Abundance
        >>> Reaction([Protein(namespace='HGNC', name='KNG1')], [Abundance(namespace='CHEBI', name='bradykinin')])

        """
        super().__init__()
        _help_named(self, namespace=namespace, identifier=identifier, name=name, xrefs=xrefs)

        if isinstance(reactants, BaseEntity):
            reactants = [reactants]
        else:
            reactants = sorted(reactants, key=_as_bel)

        if isinstance(products, BaseEntity):
            products = [products]
        else:
            products = sorted(products, key=_as_bel)

        if not reactants and not products and not namespace:
            raise ReactionEmptyException("Reaction can not be instantiated with an empty members list.")

        self.update(
            {
                REACTANTS: reactants,
                PRODUCTS: products,
            }
        )

    @property
    def reactants(self) -> List[BaseAbundance]:
        """Return the list of reactants in this reaction."""
        return self[REACTANTS]

    @property
    def products(self) -> List[BaseAbundance]:
        """Return the list of products in this reaction."""
        return self[PRODUCTS]

    def get_catalysts(self) -> Set[BaseAbundance]:
        """Get entities appearing in both the reactants and products."""
        return set(self.reactants).intersection(self.products)

    def as_bel(self, use_identifiers: bool = True) -> str:
        """Return this reaction as a BEL string."""
        return "rxn(reactants({}), products({}))".format(
            _entity_list_as_bel(self.reactants, use_identifiers=use_identifiers),
            _entity_list_as_bel(self.products, use_identifiers=use_identifiers),
        )


class ListAbundance(BaseEntity):
    """The superclass for all BEL terms defined by lists, as opposed to by names like in :class:`BaseAbundance`."""

    def __init__(self, members: Union[BaseAbundance, Iterable[BaseAbundance]]) -> None:
        """Build a list abundance node.

        :param members: A list of PyBEL node data dictionaries
        """
        super().__init__()

        if isinstance(members, BaseEntity):
            self[MEMBERS] = [members]
        else:
            self[MEMBERS] = sorted(members, key=_as_bel)

        if not self[MEMBERS]:
            raise ListAbundanceEmptyException("List abundance can not be instantiated with an empty members list.")

    @property
    def members(self) -> List[BaseAbundance]:
        """Return the list of members in this list abundance."""
        return self[MEMBERS]

    def as_bel(self, use_identifiers: bool = True) -> str:
        """Return this list abundance as a BEL string."""
        return "{}({})".format(
            self._bel_function,
            _entity_list_as_bel(self.members, use_identifiers=use_identifiers),
        )


class ComplexAbundance(ListAbundance):
    """Build a complex abundance node with the optional ability to specify a name."""

    function = COMPLEX

    def __init__(
        self,
        members: Iterable[BaseAbundance],
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        identifier: Optional[str] = None,
        xrefs: Optional[List[Entity]] = None,
    ) -> None:
        """Build a complex list node.

        :param members: A list of PyBEL node data dictionaries
        :param namespace: The namespace from which the name originates
        :param name: The name of the complex
        :param identifier: The identifier in the namespace in which the name originates
        :param xrefs: Alternate identifiers for the entity if it is named
        """
        super().__init__(members=members)
        _help_named(self, namespace=namespace, identifier=identifier, name=name, xrefs=xrefs)

    @property
    def entity(self) -> Optional[Entity]:  # noqa:D401
        """The concept represented by this complex if it has been named."""
        return self.get(CONCEPT)

    @property
    def xrefs(self) -> List[Entity]:  # noqa:D401
        """Alternative identifiers for the concept if it has been named."""
        return self.get(XREFS, [])


class NamedComplexAbundance(BaseAbundance):
    """Build a named complex abundance node.

    >>> from pybel.dsl import NamedComplexAbundance
    >>> NamedComplexAbundance(namespace='FPLX', name='Calcineurin Complex')
    """

    function = COMPLEX


class CompositeAbundance(ListAbundance):
    """Build a composite abundance node.

    This node is effectively the "AND" inside BEL, which can help
    represent when two things need to be true at the same time. For
    example, in COVID 19, if both the NF-KB and IL6-STAT complex are
    present, then acute respiratory distress syndrome happens.

    >>> from pybel.dsl import CompositeAbundance, ComplexAbundance, Protein, NamedComplexAbundance
    >>> CompositeAbundance([
    ...     NamedComplexAbundance('fplx', 'nfkb'),
    ...     ComplexAbundance([
    ...         Protein('hgnc', identifier='6018', name='IL6'),
    ...         Protein('hgnc', identifier='11364', name='STAT3'),
    ...     ]),
    ... ])
    """

    function = COMPOSITE


class FusionRangeBase(dict, metaclass=ABCMeta):
    """The superclass for fusion range data dictionaries."""

    @abstractmethod
    def as_bel(self) -> str:
        """Return this fusion range as BEL."""

    def __str__(self):  # noqa: D105
        return self.as_bel()


class MissingFusionRange(FusionRangeBase):
    """Represents a fusion range with no defined start or end."""

    def __init__(self):
        """Build a missing fusion range."""
        super(MissingFusionRange, self).__init__(
            {
                FUSION_MISSING: "?",
            }
        )

    def as_bel(self) -> str:
        """Return this missing fusion range as BEL."""
        return "?"


class EnumeratedFusionRange(FusionRangeBase):
    """Represents an enumerated fusion range."""

    def __init__(self, reference: str, start, stop):
        """Build an enumerated fusion range.

        :param reference: The reference code
        :param int or str start: The start position, either specified by its integer position, or '?'
        :param int or str stop: The stop position, either specified by its integer position, '?', or '*

        Example fully specified RNA fusion range:

        >>> EnumeratedFusionRange('r', 1, 79)

        """
        super().__init__(
            {
                FUSION_REFERENCE: reference,
                FUSION_START: start,
                FUSION_STOP: stop,
            }
        )

    def as_bel(self) -> str:
        """Return this fusion range as a BEL string."""
        return "{reference}.{start}_{stop}".format(
            reference=self[FUSION_REFERENCE],
            start=self[FUSION_START],
            stop=self[FUSION_STOP],
        )


class FusionBase(BaseEntity):
    """The superclass for building fusion node data dictionaries."""

    def __init__(
        self,
        partner_5p: CentralDogma,
        partner_3p: CentralDogma,
        range_5p: Optional[FusionRangeBase] = None,
        range_3p: Optional[FusionRangeBase] = None,
    ) -> None:
        """Build a fusion node.

        :param partner_5p: A PyBEL node for the 5-prime partner
        :param partner_3p: A PyBEL node for the 3-prime partner
        :param range_5p: A fusion range for the 5-prime partner
        :param range_3p: A fusion range for the 3-prime partner
        """
        super().__init__()
        self[FUSION] = {
            PARTNER_5P: partner_5p,
            PARTNER_3P: partner_3p,
            RANGE_5P: range_5p or MissingFusionRange(),
            RANGE_3P: range_3p or MissingFusionRange(),
        }

    @property
    def partner_5p(self) -> CentralDogma:
        """Get the 5' partner."""
        return self[FUSION][PARTNER_5P]

    @property
    def partner_3p(self) -> CentralDogma:
        """Get the 3' partner."""
        return self[FUSION][PARTNER_3P]

    @property
    def range_5p(self) -> FusionRangeBase:
        """Get the 5' partner's range."""
        return self[FUSION][RANGE_5P]

    @property
    def range_3p(self) -> FusionRangeBase:
        """Get the 3' partner's range."""
        return self[FUSION][RANGE_3P]

    def as_bel(self, use_identifiers: bool = True) -> str:
        """Return this fusion as a BEL string."""
        if use_identifiers and self.partner_3p.entity.identifier and self.partner_3p.entity.name:
            p3p = self.partner_3p.obo
        else:
            p3p = self.partner_3p.curie

        if use_identifiers and self.partner_5p.entity.identifier and self.partner_5p.entity.name:
            p5p = self.partner_5p.obo
        else:
            p5p = self.partner_5p.curie

        return '{}(fus({}, "{}", {}, "{}"))'.format(
            self._bel_function,
            p5p,
            self.range_5p.as_bel(),
            p3p,
            self.range_3p.as_bel(),
        )


class ProteinFusion(FusionBase):
    """Builds a protein fusion node."""

    function = PROTEIN


class RnaFusion(FusionBase):
    """Builds an RNA fusion node.

    Example, with fusion ranges using the 'r' qualifier:

    >>> from pybel.dsl import RnaFusion, Rna
    >>> RnaFusion(
    >>> ... partner_5p=Rna(namespace='HGNC', name='TMPRSS2'),
    >>> ... range_5p=EnumeratedFusionRange('r', 1, 79),
    >>> ... partner_3p=Rna(namespace='HGNC', name='ERG'),
    >>> ... range_3p=EnumeratedFusionRange('r', 312, 5034)
    >>> )


    Example with missing fusion ranges:

    >>> from pybel.dsl import RnaFusion, Rna
    >>> RnaFusion(
    >>> ... partner_5p=Rna(namespace='HGNC', name='TMPRSS2'),
    >>> ... partner_3p=Rna(namespace='HGNC', name='ERG'),
    >>> )

    """

    function = RNA


class GeneFusion(FusionBase):
    """Builds a gene fusion node.

    Example, using fusion ranges with the 'c' qualifier

    >>> from pybel.dsl import GeneFusion, Gene
    >>> GeneFusion(
    >>> ... partner_5p=Gene(namespace='HGNC', name='TMPRSS2'),
    >>> ... range_5p=EnumeratedFusionRange('c', 1, 79),
    >>> ... partner_3p=Gene(namespace='HGNC', name='ERG'),
    >>> ... range_3p=EnumeratedFusionRange('c', 312, 5034)
    >>> )

    Example with missing fusion ranges:

    >>> from pybel.dsl import GeneFusion, Gene
    >>> GeneFusion(
    >>> ... partner_5p=Gene(namespace='HGNC', name='TMPRSS2'),
    >>> ... partner_3p=Gene(namespace='HGNC', name='ERG'),
    >>> )

    """

    function = GENE



# -*- coding: utf-8 -*-

"""This module contains IO functions for BEL scripts."""

import gzip
import logging
from typing import TextIO, Union

from bel_resources.utils import download
from networkx.utils import open_file

from .line_utils import parse_lines
from ..struct import BELGraph

__all__ = [
    "from_bel_script",
    "from_bel_script_gz",
    "from_bel_script_url",
]

logger = logging.getLogger(__name__)


@open_file(0, mode="r")
def from_bel_script(path: Union[str, TextIO], **kwargs) -> BELGraph:
    """Load a BEL graph from a file resource. This function is a thin wrapper around :func:`from_lines`.

    :param path: A path or file-like

    The remaining keyword arguments are passed to :func:`pybel.io.line_utils.parse_lines`,
    which populates a :class:`BELGraph`.
    """
    logger.info("Reading BEL script at %s", path.name)
    graph = BELGraph(path=path.name)
    parse_lines(graph=graph, lines=path, **kwargs)
    return graph


def from_bel_script_gz(path, **kwargs) -> BELGraph:
    """Parse a BEL graph from a gzipped BEL Script."""
    with gzip.open(path, "rt") as file:
        return from_bel_script(file, **kwargs)


def from_bel_script_url(url: str, **kwargs) -> BELGraph:
    """Load a BEL graph from a URL resource.

    :param url: A valid URL pointing to a BEL document

    The remaining keyword arguments are passed to :func:`pybel.io.line_utils.parse_lines`.
    """
    logger.info("Loading from url: %s", url)
    res = download(url)
    lines = (line.decode("utf-8") for line in res.iter_lines())

    graph = BELGraph(path=url)
    parse_lines(graph=graph, lines=lines, **kwargs)
    return graph



# -*- coding: utf-8 -*-

"""Utilities for PyBEL testing."""

import itertools as itt
import random

from .utils import n
from ..dsl import protein
from ..struct import BELGraph

__all__ = [
    "generate_random_graph",
]


def generate_random_graph(n_nodes, n_edges, namespace="NS"):
    """Generate a sub-graph with random nodes and edges.

    :param int n_nodes: Number of nodes to make
    :param int n_edges: Number of edges to make
    :param str namespace: The namespace of the nodes to use
    :rtype: pybel.BELGraph
    """
    graph = BELGraph()

    nodes = [protein(namespace=namespace, name=str(i)) for i in range(1, n_nodes)]

    edges = list(itt.combinations(nodes, r=2))
    edge_sample = random.sample(edges, n_edges)

    for u, v in edge_sample:
        graph.add_increases(u, v, citation=n(), evidence=n())

    return graph



# -*- coding: utf-8 -*-

"""Convert BEL graphs to HTML.

This module provides functions for making HTML visualizations of BEL Graphs. Because the :class:`pybel.BELGraph`
inherits from :class:`networkx.MultiDiGraph`, it can also be visualized using :mod:`networkx`
`library <https://networkx.github.io/documentation/latest/reference/drawing.html>`_.
"""

import json
from typing import Mapping, Optional, TextIO, Union

from networkx.utils import open_file

from .constants import DEFAULT_COLOR_MAP
from ..jinja_utils import build_template_renderer
from ..nodelink import to_nodelink_jsons
from ...struct import BELGraph

__all__ = [
    "to_html",
    "to_html_file",
]


def to_html(graph: BELGraph, color_map: Optional[Mapping[str, str]] = None) -> str:
    """Create an HTML visualization for the given JSON representation of a BEL graph.

    :param graph: A BEL graph
    :param color_map: A dictionary from PyBEL internal node functions to CSS color strings like #FFEE00.
                    Defaults to :data:`default_color_map`
    :return: HTML string representing the graph
    """
    color_map = DEFAULT_COLOR_MAP if color_map is None else color_map
    render_template = build_template_renderer(__file__)
    return render_template(
        "graph_template.html",
        graph=to_nodelink_jsons(graph),
        color_map=json.dumps(color_map),
        number_nodes=graph.number_of_nodes(),
        number_edges=graph.number_of_edges(),
    )


@open_file(1, mode="w")
def to_html_file(
    graph: BELGraph,
    file: Union[str, TextIO],
    color_map: Optional[Mapping[str, str]] = None,
) -> None:
    """Write the HTML visualization to a file or file-like.

    :param graph: A BEL graph
    :param color_map: A dictionary from PyBEL internal node functions to CSS color strings like #FFEE00.
                    Defaults to :data:`default_color_map`
    :param file file: A writable file or file-like or file path
    """
    print(to_html(graph, color_map=color_map), file=file)



# -*- coding: utf-8 -*-

"""Constants for PyBEL-Jupyter."""

from ...constants import (
    ABUNDANCE,
    BIOPROCESS,
    COMPLEX,
    COMPOSITE,
    GENE,
    MIRNA,
    PATHOLOGY,
    PROTEIN,
    REACTION,
    RNA,
)

#: The color map defining the node colors in visualization
DEFAULT_COLOR_MAP = {
    PROTEIN: "#1F77B4",
    PATHOLOGY: "#FF7F0E",
    BIOPROCESS: "#2CA02C",
    MIRNA: "#D62728",
    COMPLEX: "#98DF8A",
    COMPOSITE: "#9467BD",
    REACTION: "#000000",
    GENE: "#FFBB78",
    ABUNDANCE: "#AEC7E8",
    RNA: "#FF9896",
}



# -*- coding: utf-8 -*-

"""Functions for enriching the origins of Proteins, RNAs, and miRNAs."""

from ...graph import BELGraph
from ...pipeline import in_place_transformation
from ....constants import FUNCTION, FUSION, MIRNA, RNA, VARIANTS
from ....dsl import Protein

__all__ = [
    "enrich_rnas_with_genes",
    "enrich_proteins_with_rnas",
    "enrich_protein_and_rna_origins",
]


@in_place_transformation
def enrich_proteins_with_rnas(graph: BELGraph) -> None:
    """Add the corresponding RNA node for each protein node and connect them with a translation edge.

    :param graph: A BEL graph
    """
    for protein_node in list(graph):
        if not isinstance(protein_node, Protein):
            continue

        if protein_node.variants:
            continue

        rna_node = protein_node.get_rna()
        graph.add_translation(rna_node, protein_node)


@in_place_transformation
def enrich_rnas_with_genes(graph: BELGraph) -> None:
    """Add the corresponding gene node for each RNA/miRNA node and connect them with a transcription edge.

    :param graph: A BEL graph
    """
    for rna_node in list(graph):
        if rna_node[FUNCTION] not in {MIRNA, RNA} or FUSION in rna_node or VARIANTS in rna_node:
            continue

        gene_node = rna_node.get_gene()
        graph.add_transcription(gene_node, rna_node)


@in_place_transformation
def enrich_protein_and_rna_origins(graph: BELGraph) -> None:
    """Add the corresponding RNA for each protein then the corresponding gene for each RNA/miRNA.

    :param graph: A BEL graph
    """
    enrich_proteins_with_rnas(graph)
    enrich_rnas_with_genes(graph)



# -*- coding: utf-8 -*-

"""An example describing a single evidence about BRAF.

.. code-block:: none

    SET Citation = {"PubMed", "11283246"}
    SET Evidence = "Expression of both dominant negative forms, RasN17 and Rap1N17, in UT7-Mpl cells decreased
    thrombopoietin-mediated Elk1-dependent transcription. This suggests that both Ras and Rap1 contribute to
    thrombopoietin-induced ELK1 transcription."

    SET Species = 9606

    p(HGNC:THPO) increases kin(p(HGNC:BRAF))
    p(HGNC:THPO) increases kin(p(HGNC:RAF1))
    kin(p(HGNC:BRAF)) increases tscript(p(HGNC:ELK1))

    UNSET ALL
"""

from ..dsl import Entity, Protein, activity
from ..resources import HGNC_URL, SPECIES_PATTERN
from ..struct.graph import BELGraph

__all__ = [
    "braf_graph",
]

braf_graph = BELGraph(
    name="BRAF Subgraph",
    version="1.0.0",
    description="Some relations surrounding BRAF",
    authors="Charles Tapley Hoyt",
    contact="cthoyt@gmail.com",
)

braf_graph.namespace_url.update(
    {
        "HGNC": HGNC_URL,
    }
)

braf_graph.annotation_pattern.update(
    {
        "Species": SPECIES_PATTERN,
    }
)

thpo = Protein(
    namespace="HGNC",
    name="THPO",
    identifier="11795",
    xrefs=[
        Entity(namespace="uniprot", identifier="P40225"),
    ],
)
braf = Protein(namespace="HGNC", name="BRAF", identifier="1097")
raf1 = Protein(namespace="HGNC", name="RAF1", identifier="9829")
elk1 = Protein(namespace="HGNC", name="ELK1", identifier="3321")

evidence = (
    "Expression of both dominant negative forms, RasN17 and Rap1N17, in UT7-Mpl cells decreased "
    "thrombopoietin-mediated Elk1-dependent transcription. This suggests that both Ras and Rap1 contribute to "
    "thrombopoietin-induced ELK1 transcription."
)

braf_graph.add_increases(
    thpo,
    braf,
    evidence=evidence,
    citation="11283246",
    target_modifier=activity(name="kin"),
    annotations={"Species": "9606"},
)

braf_graph.add_increases(
    thpo,
    raf1,
    evidence=evidence,
    citation="11283246",
    target_modifier=activity(name="kin"),
    annotations={"Species": "9606"},
)

braf_graph.add_increases(
    braf,
    elk1,
    evidence=evidence,
    citation="11283246",
    source_modifier=activity(name="kin"),
    target_modifier=activity(name="tscript"),
    annotations={"Species": "9606"},
)



# -*- coding: utf-8 -*-

"""This module facilitates the transfer of knowledge through ontological relationships."""

from typing import Iterable, List

from ...graph import BELGraph
from ....constants import (
    ANNOTATIONS,
    CAUSAL_RELATIONS,
    CITATION,
    EVIDENCE,
    IS_A,
    RELATION,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
)
from ....dsl import BaseEntity

__all__ = [
    "infer_child_relations",
]


def iter_children(graph: BELGraph, node: BaseEntity) -> Iterable[BaseEntity]:
    """Iterate over children of the node."""
    return (node for node, _, d in graph.in_edges(node, data=True) if d[RELATION] == IS_A)


def transfer_causal_edges(graph: BELGraph, source: BaseEntity, target: BaseEntity) -> Iterable[str]:
    """Transfer causal edges that the source has to the target and yield the resulting hashes."""
    for _, v, data in graph.out_edges(source, data=True):
        if data[RELATION] not in CAUSAL_RELATIONS:
            continue

        yield graph.add_qualified_edge(
            target,
            v,
            relation=data[RELATION],
            evidence=data[EVIDENCE],
            citation=data[CITATION],
            annotations=data.get(ANNOTATIONS),
            source_modifier=data.get(SOURCE_MODIFIER),
            target_modifier=data.get(TARGET_MODIFIER),
        )

    for u, _, data in graph.in_edges(source, data=True):
        if data[RELATION] not in CAUSAL_RELATIONS:
            continue

        yield graph.add_qualified_edge(
            u,
            target,
            relation=data[RELATION],
            evidence=data[EVIDENCE],
            citation=data[CITATION],
            annotations=data.get(ANNOTATIONS),
            source_modifier=data.get(SOURCE_MODIFIER),
            target_modifier=data.get(TARGET_MODIFIER),
        )


def infer_child_relations(graph: BELGraph, node: BaseEntity) -> List[str]:
    """Propagate causal relations to children."""
    return list(_infer_child_relations_iter(graph, node))


def _infer_child_relations_iter(graph: BELGraph, node: BaseEntity) -> Iterable[str]:
    """Propagate causal relations to children."""
    for child in iter_children(graph, node):
        yield from transfer_causal_edges(graph, node, child)
        yield from infer_child_relations(graph, child)



# -*- coding: utf-8 -*-

"""TSV converter classes."""

from abc import ABC, abstractmethod
from typing import Dict, Tuple

from ...constants import (
    ACTIVITY,
    ASSOCIATION,
    CAUSAL_DECREASE_RELATIONS,
    CAUSAL_INCREASE_RELATIONS,
    CAUSAL_RELATIONS,
    CAUSES_NO_CHANGE,
    CORRELATIVE_RELATIONS,
    DECREASES,
    DEGRADATION,
    DIRECTLY_DECREASES,
    DIRECTLY_INCREASES,
    EQUIVALENT_TO,
    HAS_PRODUCT,
    HAS_REACTANT,
    HAS_VARIANT,
    INCREASES,
    IS_A,
    MODIFIER,
    PART_OF,
    REGULATES,
    RELATION,
    TARGET_MODIFIER,
)
from ...dsl import (
    Abundance,
    BaseAbundance,
    BaseEntity,
    BiologicalProcess,
    CentralDogma,
    ComplexAbundance,
    Gene,
    MicroRna,
    NamedComplexAbundance,
    Pathology,
    Population,
    Protein,
    Reaction,
    Rna,
)
from ...typing import EdgeData


def _safe_label(base_entity: BaseEntity):
    return base_entity.safe_label


class Converter(ABC):
    """A condition and converter for a BEL edge."""

    @staticmethod
    @abstractmethod
    def predicate(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""

    @staticmethod
    @abstractmethod
    def convert(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> Tuple[str, str, str]:
        """Convert a BEL edge."""


class SimpleConverter(Converter):
    """A class for converting the source and target that have simple names."""

    @classmethod
    def convert(cls, u: BaseAbundance, v: BaseAbundance, key: str, edge_data: EdgeData) -> Tuple[str, str, str]:
        """Convert a BEL edge."""
        return u.safe_label, edge_data[RELATION], v.safe_label


class TypedConverter(Converter):
    """A class for converting the source and target but replaces the relation."""

    target_relation = None

    @classmethod
    def convert(cls, u: BaseAbundance, v: BaseAbundance, key: str, edge_data: EdgeData) -> Tuple[str, str, str]:
        """Convert a BEL edge."""
        return u.safe_label, cls.target_relation, v.safe_label


class SimplePredicate(Converter):
    """Converts BEL statements based on a given relation."""

    relation = ...

    @classmethod
    def predicate(cls, u, v, key, edge_data) -> bool:
        """Test a BEL edge has a given relation."""
        return edge_data[RELATION] == cls.relation


class SimpleTypedPredicate(SimplePredicate):
    """Finds BEL statements like ``A(X) B C(Y)`` where relation B and types A and C are defined in the class."""

    subject_type = ...
    object_type = ...

    @classmethod
    def predicate(cls, u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        return (
            super().predicate(u, v, key, edge_data)
            and isinstance(u, cls.subject_type)
            and isinstance(v, cls.object_type)
        )


class HasVariantConverter(SimpleConverter, SimpleTypedPredicate):
    """Converts BEL statements like ``p(X) hasVariant p(X, ...)``."""

    subject_type = CentralDogma
    relation = HAS_VARIANT
    object_type = CentralDogma


class _PartOfConverter(SimpleTypedPredicate, TypedConverter):
    relation = PART_OF
    target_relation = "partOf"


class PartOfNamedComplexConverter(_PartOfConverter):
    """Converts BEL statements like ``p(X) partOf complex(Y)``."""

    subject_type = Protein
    object_type = NamedComplexAbundance


class ProcessCausalConverter(SimpleConverter, SimpleTypedPredicate):
    """Converts BEL statements like ``bp(X) increases/decreases bp(Y)``."""

    subject_type = BiologicalProcess
    relations = CAUSAL_RELATIONS
    object_type = BiologicalProcess

    @classmethod
    def predicate(cls, u, v, key, edge_data) -> bool:
        """Test a BEL edge has a given relation."""
        return (
            isinstance(u, cls.subject_type) and edge_data[RELATION] in cls.relations and isinstance(v, cls.object_type)
        )


class SubprocessPartOfBiologicalProcessConverter(_PartOfConverter):
    """Converts BEL statements like ``bp(X) partOf bp(Y)``."""

    subject_type = BiologicalProcess
    object_type = BiologicalProcess


class ProteinPartOfBiologicalProcessConverter(_PartOfConverter):
    """Converts BEL statements like ``p(X) partOf bp(Y)``."""

    subject_type = Protein
    object_type = BiologicalProcess


class AbundancePartOfPopulationConverter(_PartOfConverter):
    """Converts BEL statements like ``a(X) partOf pop(Y)``."""

    subject_type = Abundance
    object_type = Population


class PopulationPartOfAbundanceConverter(_PartOfConverter):
    """Converts BEL statements like ``pop(X) partOf a(Y)``."""

    subject_type = Population
    object_type = Abundance


class _ReactionTypedPredicate(SimpleTypedPredicate):
    subject_type = Reaction
    object_type = BaseAbundance


class _ReactionHasMemberConverter(_ReactionTypedPredicate):
    """Converts BEL statements like ``complex(X) hasComponent p(Y)``."""

    target_relation = ...

    @classmethod
    def predicate(cls, u: Reaction, v: BaseAbundance, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        return super().predicate(u, v, key, edge_data) and v not in u.get_catalysts()

    @classmethod
    def convert(cls, u: Reaction, v: BaseAbundance, key: str, data: Dict) -> Tuple[str, str, str]:
        """Convert a BEL edge."""
        return u.as_bel(), cls.target_relation, v.curie


class ReactionHasReactantConverter(_ReactionHasMemberConverter):
    """Converts BEL statements like ``rxn(X) hasReactant a(Y)``."""

    relation = HAS_REACTANT
    target_relation = "hasReactant"


class ReactionHasProductConverter(_ReactionHasMemberConverter):
    """Converts BEL statements like ``rxn(X) hasProduct a(Y)``."""

    relation = HAS_PRODUCT
    target_relation = "hasProduct"


class ReactionHasCatalystConverter(_ReactionTypedPredicate):
    """Converts BEL statements that simultaneously ``rxn(X) hasProduct a(Y)`` and ``rxn(X) hasReactant a(Y)``."""

    target_relation = "hasCatalyst"

    @classmethod
    def predicate(cls, u: Reaction, v: BaseAbundance, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        return super().predicate(u, v, key, edge_data) and v in u.get_catalysts()

    @classmethod
    def convert(cls, u: Reaction, v: BaseAbundance, key: str, data: Dict) -> Tuple[str, str, str]:
        """Convert a BEL edge."""
        return u.as_bel(), cls.target_relation, v.curie


class ListComplexHasComponentConverter(SimpleTypedPredicate):
    """Converts BEL statements like ``complex(p(X), p(Y), ...) hasComponent p(X)``."""

    subject_type = BaseAbundance
    relation = PART_OF
    object_type = ComplexAbundance
    target_relation = "partOf"

    @classmethod
    def convert(cls, u: ComplexAbundance, v: BaseAbundance, key: str, data: Dict) -> Tuple[str, str, str]:
        """Convert a BEL edge."""
        return u.curie, cls.target_relation, v.as_bel()


class IsAConverter(SimplePredicate, SimpleConverter):
    """Converts BEL statements like ``X isA Y``."""

    relation = IS_A
    target_relation = "isA"


class EquivalenceConverter(SimplePredicate, SimpleConverter):
    """Converts BEL statements like ``X eq Y``."""

    relation = EQUIVALENT_TO
    target_relation = "equivalentTo"


class CorrelationConverter(SimpleConverter):
    """Converts BEL statements like ``A(B) pos|neg|noCorrelation C(D)``."""

    @staticmethod
    def predicate(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        return edge_data[RELATION] in CORRELATIVE_RELATIONS


class AssociationConverter(Converter):
    """Converts BEL statements like ``a(X) -- path(Y)``."""

    @staticmethod
    def predicate(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        return edge_data[RELATION] == ASSOCIATION

    @staticmethod
    def convert(u: BaseAbundance, v: BaseAbundance, key: str, edge_data: EdgeData) -> Tuple[str, str, str]:
        """Convert a BEL edge."""
        relation = edge_data.get("association_type", ASSOCIATION)  # allow more specific association to be defined
        return u.safe_label, relation, v.safe_label


class DrugEffectConverter(SimpleConverter, SimpleTypedPredicate):
    """Converts BEL statements like ``a(X) ? path(Y)``."""

    subject_type = Abundance
    relation = ...
    object_type = Pathology


class DrugIndicationConverter(DrugEffectConverter):
    """Converts BEL statements like ``a(X) -| path(Y)``."""

    relation = DECREASES


class DrugSideEffectConverter(DrugEffectConverter):
    """Converts BEL statements like ``a(X) -> path(Y)``."""

    relation = INCREASES


class RegulatesAmountConverter(TypedConverter):
    """Converts BEL statements like ``A(B) reg C(D)``."""

    relation = REGULATES
    target_relation = "regulatesAmountOf"

    @classmethod
    def predicate(cls, u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        target_modifier = edge_data.get(TARGET_MODIFIER)
        return edge_data[RELATION] == cls.relation and (not target_modifier or not target_modifier.get(MODIFIER))


class IncreasesAmountConverter(RegulatesAmountConverter):
    """Converts BEL statements like ``A(B) -> C(D)``."""

    relation = INCREASES
    target_relation = "increasesAmountOf"


class DecreasesAmountConverter(RegulatesAmountConverter):
    """Converts BEL statements like ``A(B) -| C(D)``."""

    relation = DECREASES
    target_relation = "decreasesAmountOf"


class NoChangeAmountConverter(RegulatesAmountConverter):
    """Converts BEL statements like ``A(B) cnc C(D)``."""

    relation = CAUSES_NO_CHANGE
    target_relation = "notRegulatesAmountOf"


class RegulatesDegradationConverter(TypedConverter):
    """Converts BEL statements like ``A(B) reg deg(C(D))``."""

    relation = REGULATES
    target_relation = "regulatesAmountOf"

    @classmethod
    def predicate(cls, u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        target_modifier = edge_data.get(TARGET_MODIFIER)
        return edge_data[RELATION] == cls.relation and target_modifier and target_modifier.get(MODIFIER) == DEGRADATION


class IncreasesDegradationConverter(RegulatesDegradationConverter):
    """Converts BEL statements like ``A(B) -> deg(C(D))``."""

    relation = INCREASES
    target_relation = "decreasesAmountOf"


class DecreasesDegradationConverter(RegulatesDegradationConverter):
    """Converts BEL statements like ``A(B) -| deg(C(D))``."""

    relation = DECREASES
    target_relation = "increasesAmountOf"


class NoChangeDegradationConverter(RegulatesDegradationConverter):
    """Converts BEL statements like ``A(B) cnc deg(C(D))``."""

    relation = CAUSES_NO_CHANGE
    target_relation = "notRegulatesAmountOf"


class RegulatesActivityConverter(TypedConverter):
    """Converts BEL statements like ``A(B) reg act(C(D) [, ma(E)])``."""

    relation = REGULATES
    target_relation = "activityDirectlyRegulatesActivityOf"

    @classmethod
    def predicate(cls, u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        target_modifier = edge_data.get(TARGET_MODIFIER)
        return edge_data[RELATION] == cls.relation and target_modifier and target_modifier.get(MODIFIER) == ACTIVITY


class IncreasesActivityConverter(RegulatesActivityConverter):
    """Converts BEL statements like ``A(B) -> act(C(D) [, ma(E)])``."""

    relation = INCREASES
    target_relation = "activityPositivelyRegulatesActivityOf"


class DirectlyIncreasesActivityConverter(RegulatesActivityConverter):
    """Converts BEL statements like ``A(B) => act(C(D) [, ma(E)])``."""

    relation = DIRECTLY_INCREASES
    target_relation = "activityDirectlyPositivelyRegulatesActivityOf"


class DecreasesActivityConverter(RegulatesActivityConverter):
    """Converts BEL statements like ``A(B) -| act(C(D) [, ma(E)])``."""

    relation = DECREASES
    target_relation = "activityNegativelyRegulatesActivityOf"


class DirectlyDecreasesActivityConverter(RegulatesActivityConverter):
    """Converts BEL statements like ``A(B) =| act(C(D) [, ma(E)])``."""

    relation = DIRECTLY_DECREASES
    target_relation = "activityDirectlyNegativelyRegulatesActivityOf"


class NoChangeActivityConverter(RegulatesActivityConverter):
    """Converts BEL statements like ``A(B) cnc act(C(D) [, ma(E)])``."""

    relation = CAUSES_NO_CHANGE
    target_relation = "notActivityDirectlyRegulatesActivityOf"


class AbundanceDirectlyDecreasesProteinActivityConverter(DirectlyDecreasesActivityConverter):
    """Converts BEL statements like ``a(X) =| act(p(Y))``."""

    subject_type = Abundance
    object_type = Protein


class AbundanceDirectlyIncreasesProteinActivityConverter(DirectlyIncreasesActivityConverter):
    """Converts BEL statements like ``a(X) => act(p(Y))``."""

    subject_type = Abundance
    object_type = Protein


class MiRNARegulatesExpressionConverter(TypedConverter, SimpleTypedPredicate):
    """Converts BEL statements like ``m(X) reg r(Y)``."""

    subject_type = MicroRna
    relation = REGULATES
    object_type = Rna
    target_relation = "regulatesExpressionOf"


class MiRNAIncreasesExpressionConverter(MiRNARegulatesExpressionConverter):
    """Converts BEL statements like ``m(X) -> r(Y)``."""

    relation = INCREASES
    target_relation = "increasesExpressionOf"


class MiRNADirectlyIncreasesExpressionConverter(MiRNARegulatesExpressionConverter):
    """Converts BEL statements like ``m(X) => r(Y)``."""

    relation = DIRECTLY_INCREASES
    target_relation = "increasesExpressionOf"


class MiRNADecreasesExpressionConverter(MiRNARegulatesExpressionConverter):
    """Converts BEL statements like ``m(X) -| r(Y)``."""

    relation = DECREASES
    target_relation = "repressesExpressionOf"


class MiRNADirectlyDecreasesExpressionConverter(MiRNARegulatesExpressionConverter):
    """Converts BEL statements like ``m(X) =| r(Y)``."""

    relation = DIRECTLY_DECREASES
    target_relation = "repressesExpressionOf"


class TranscriptionFactorForConverter(Converter):
    """Converts ``complex(g(A), p(B)) directlyIncreases r(A)```."""

    @classmethod
    def convert(cls, u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> Tuple[str, str, str]:
        """Convert a transcription factor for edge."""
        gene = v.get_gene()
        if gene == u.members[0]:
            return u.members[1].safe_label, edge_data[RELATION], v.safe_label
        else:
            return u.members[0].safe_label, edge_data[RELATION], v.safe_label

    @classmethod
    def predicate(cls, u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        if not isinstance(u, ComplexAbundance) or len(u.members) != 2:
            return False

        if isinstance(u.members[0], Gene) and isinstance(u.members[1], Protein):
            gene = u.members[0]
        elif isinstance(u.members[1], Gene) and isinstance(u.members[0], Protein):
            gene = u.members[1]
        else:
            return False

        if not isinstance(v, Rna):
            return False

        return gene == v.get_gene()


class BindsProteinConverter(Converter):
    """Converts ``x(B) => complex(p(A), x(B))```."""

    @staticmethod
    def predicate(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        return (
            edge_data[RELATION] == DIRECTLY_INCREASES
            and isinstance(v, ComplexAbundance)
            and len(v.members) == 2
            and u in v.members
            and isinstance([m for m in v.members if m != u][0], Protein)
        )

    @staticmethod
    def convert(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> Tuple[str, str, str]:
        """Convert a binds protein factor for edge."""
        v = [m for m in v.members if m != u][0]
        return u.safe_label, "bindsToProtein", v.safe_label


class HomomultimerConverter(Converter):
    """Converts ``p(A) directlyIncreases complex(p(A), p(A))```."""

    @staticmethod
    def predicate(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        return (
            isinstance(u, Protein)
            and edge_data[RELATION] == DIRECTLY_INCREASES
            and isinstance(v, ComplexAbundance)
            and all(member == u for member in v.members)
        )

    @staticmethod
    def convert(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> Tuple[str, str, str]:
        """Convert a homomultimer formation."""
        return u.safe_label, "bindsToProtein", u.safe_label


class BindsGeneConverter(Converter):
    """Converts ``p(B) directlyIncreases complex(g(A), p(B))```."""

    @staticmethod
    def predicate(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        return (
            isinstance(u, Protein)
            and edge_data[RELATION] == DIRECTLY_INCREASES
            and isinstance(v, ComplexAbundance)
            and len(v.members) == 2
            and u in v.members
            and isinstance([m for m in v.members if m != u][0], Gene)
        )

    @staticmethod
    def convert(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> Tuple[str, str, str]:
        """Convert a transcription factor for edge."""
        v = [m for m in v.members if m != u][0]
        return u.safe_label, "bindsToGene", v.safe_label


class ProteinRegulatesComplex(Converter):
    """Converts ``p(B) directlyIncreases complex(x(X), y(Y))```."""

    @staticmethod
    def predicate(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> bool:
        """Test a BEL edge."""
        return (
            isinstance(u, Protein)
            and isinstance(v, ComplexAbundance)
            and u not in v.members
            and edge_data[RELATION] in CAUSAL_RELATIONS
            and edge_data[RELATION] != CAUSES_NO_CHANGE
        )

    @staticmethod
    def convert(u: BaseEntity, v: BaseEntity, key: str, edge_data: EdgeData) -> Tuple[str, str, str]:
        """Convert a transcription factor for edge."""
        relation = edge_data[RELATION]
        if relation in CAUSAL_INCREASE_RELATIONS:
            relation = "increasesAmountOf"
        elif relation in CAUSAL_DECREASE_RELATIONS:
            relation = "decreasesAmountOf"
        elif relation == REGULATES:
            relation = "regulatesAmountOf"
        else:
            raise ValueError("invalid relation type")

        return u.safe_label, relation, v.safe_label



# -*- coding: utf-8 -*-

"""This module contains IO functions for outputting BEL graphs to lossy formats, such as GraphML and CSV."""

import json
from typing import Optional, TextIO, Union

from networkx.utils import open_file

from ..dsl import CentralDogma
from ..struct import BELGraph

__all__ = [
    "to_csv",
    "to_sif",
    "to_gsea",
]


@open_file(1, mode="w")
def to_csv(graph: BELGraph, path: Union[str, TextIO], sep: Optional[str] = None) -> None:
    """Write the graph as a tab-separated edge list.

    The resulting file will contain the following columns:

    1. Source BEL term
    2. Relation
    3. Target BEL term
    4. Edge data dictionary

    See the Data Models section of the documentation for which data are stored in the edge data dictionary, such
    as queryable information about transforms on the subject and object and their associated metadata.
    """
    if sep is None:
        sep = "\t"

    for u, v, data in graph.edges(data=True):
        print(
            graph.edge_to_bel(u, v, edge_data=data, sep=sep),
            json.dumps(data),
            sep=sep,
            file=path,
        )


@open_file(1, mode="w")
def to_sif(graph: BELGraph, path: Union[str, TextIO], sep: Optional[str] = None) -> None:
    """Write the graph as a tab-separated SIF file.

    The resulting file will contain the following columns:

    1. Source BEL term
    2. Relation
    3. Target BEL term

    This format is simple and can be used readily with many applications, but is lossy in that it does not include
    relation metadata.
    """
    if sep is None:
        sep = "\t"

    for u, v, data in graph.edges(data=True):
        print(
            graph.edge_to_bel(u, v, edge_data=data, sep=sep),
            file=path,
        )


@open_file(1, mode="w")
def to_gsea(graph: BELGraph, path: Union[str, TextIO]) -> None:
    """Write the genes/gene products to a GRP file for use with GSEA gene set enrichment analysis.

    .. seealso::

        - GRP `format specification <http://software.broadinstitute.org/cancer/software/gsea/wiki/index.php/Data_formats#GRP:_Gene_set_file_format_.28.2A.grp.29>`_
        - GSEA `publication <https://doi.org/10.1073/pnas.0506580102>`_
    """
    print("# {}".format(graph.name), file=path)
    hgnc_gene_symbols = {
        node.name for node in graph if isinstance(node, CentralDogma) and node.namespace.lower() == "hgnc"
    }
    for hgnc_gene_symbol in sorted(hgnc_gene_symbols):
        print(hgnc_gene_symbol, file=path)



# -*- coding: utf-8 -*-

"""Tests for instantiating the manager"""

import os
import tempfile
import unittest

from pybel import Manager
from pybel.manager.base_manager import build_engine_session

try:
    from unittest import mock
except ImportError:
    import mock


class TestInstantiation(unittest.TestCase):
    """Allows for testing with a consistent connection without changing the configuration."""

    def setUp(self):
        """Add two class-level variables: ``mock_global_connection`` and ``mock_module_connection`` that can be
        used as context managers to mock the bio2bel connection getter functions."""

        self.fd, self.path = tempfile.mkstemp()
        self.connection = "sqlite:///" + self.path

        def mock_connection():
            """Get the connection enclosed by this class.

            :rtype: str
            """
            return self.connection

        self.mock_connection = mock.patch("pybel.manager.cache_manager.get_cache_connection", mock_connection)

    def tearDown(self):
        os.close(self.fd)
        os.remove(self.path)

    def test_fail_connection_none(self):
        """Test that a None causes a huge error."""
        with self.assertRaises(ValueError):
            build_engine_session(None)

    def test_instantiate_init(self):
        """Test what happens when no connection is specified for the normal constructor."""
        with self.mock_connection:
            manager = Manager()
            self.assertEqual(self.connection, str(manager.engine.url))

    def test_instantiate_manager_positional(self):
        manager = Manager(self.connection)
        self.assertEqual(self.connection, str(manager.engine.url))

    def test_instantiate_manager_positional_with_keyword(self):
        manager = Manager(self.connection, echo=False)
        self.assertEqual(self.connection, str(manager.engine.url))

    def test_instantiate_manager_fail_positional(self):
        with self.assertRaises(ValueError):
            Manager(self.connection, True)

    def test_instantiate_manager_keyword(self):
        manager = Manager(connection=self.connection)
        self.assertEqual(self.connection, str(manager.engine.url))

    def test_instantiate_manager_connection_fail_too_many_keyword(self):
        with self.assertRaises(ValueError):
            Manager(connection=self.connection, engine="something", session="something")

    def test_instantiate_manager_engine_fail_too_many_keywords(self):
        with self.assertRaises(ValueError):
            Manager(engine="something", session="something", echo=False)

    def test_instantiate_manager_engine_missing(self):
        with self.assertRaises(ValueError):
            Manager(engine=None, session="fake-session")

    def test_instantiate_manager_session_missing(self):
        with self.assertRaises(ValueError):
            Manager(engine="fake-engine", session=None)



# -*- coding: utf-8 -*-

"""Convert a BEL graph to HiPathia inputs.

Input
-----
SIF File
~~~~~~~~
- Text file with three columns separated by tabs.
- Each row represents an interaction in the pathway. First column is the source
  node, third column the target node, and the second is the type of relation
  between them.
- Only activation and inhibition interactions are allowed.
- The name of the nodes in this file will be stored as the IDs of the nodes.
- The nodes IDs should have the following structure: N (dash) pathway ID (dash)
  node ID.
- HiPathia distinguish between two types of nodes: simple and complex.

Simple nodes:

- Simple nodes may include many genes, but only one is needed to perform the
  function of the node. This could correspond to a protein family of enzymes
  that all have the same function - only one of them needs to be present for
  the action to take place. Simple nodes are defined within
- Node IDs from simple nodes do not include any space, i.e. N-hsa04370-11.

Complex nodes:

- Complex nodes include different simple nodes and represent protein complexes.
  Each simple node within the complex represents one protein in the complex.
  This node requires the presence of all their simple nodes to perform its
  function.
- Node IDs from complex nodes are the juxtaposition of the included simple node
  IDs, separated by spaces, i.e. N-hsa04370-10 26.

ATT File
~~~~~~~~
Text file with twelve (12) columns separated by tabulars. Each row represents a node (either simple or complex).

The columns included are:

1. ``ID``: Node ID as explained above.
2. ``label``: Name to be shown in the picture of the pathway en HGNC. Generally, the gene name of the first
   included EntrezID gene is used as label. For complex nodes, we juxtapose the gene names of the first genes of
   each simple node included (see genesList column below).
3. ``X``: The X-coordinate of the position of the node in the pathway.
4. ``Y``: The Y-coordinate of the position of the node in the pathway.
5. ``color``: The default color of the node.
6. ``shape``: The shape of the node. "rectangle" should be used for genes and "circle" for metabolites.
7. ``type``: The type of the node, either "gene" for genes or "compound" for metabolites. For complex nodes, the
   type of each of their included simple nodes is juxtaposed separated by commas, i.e. gene,gene.
8. ``label.cex``: Amount by which plotting label should be scaled relative to the default.
9. ``label.color``: Default color of the node.
10. ``width``: Default width of the node.
11. ``height``: Default height of the node.
12. ``genesList``: List of genes included in each node, with EntrezID:

  - Simple nodes: EntrezIDs of the genes included, separated by commas (",") and no spaces, i.e. 56848,8877 for
    node N-hsa04370-11.
  - Complex nodes: GenesList of the simple nodes included, separated by a slash ("/") and no spaces, and in the
    same order as in the node ID. For example, node N-hsa04370-10 26 includes two simple nodes: 10 and 26. Its
    genesList column is 5335,5336,/,9047, meaning that the genes included in node 10 are 5335 and 5336, and the
    gene included in node 26 is 9047.

"""

import logging
import os
from collections import defaultdict
from itertools import groupby
from operator import itemgetter
from typing import List, Optional, Set, Tuple, Union

import networkx as nx
import pandas as pd

from ..constants import (
    CAUSAL_INCREASE_RELATIONS,
    CAUSAL_POLAR_RELATIONS,
    CITATION_TYPE_OTHER,
    IS_A,
    RELATION,
)
from ..dsl import ComplexAbundance, Protein, hgnc
from ..struct import BELGraph

__all__ = [
    "from_hipathia_paths",
    "from_hipathia_dfs",
    "to_hipathia",
    "to_hipathia_dfs",
]

logger = logging.getLogger(__name__)

ATT_COLS = ["ID", "label", "genesList"]


def from_hipathia_paths(name: str, att_path: str, sif_path: str) -> BELGraph:
    """Get a BEL graph from HiPathia files."""
    att_df = pd.read_csv(att_path, sep="\t")
    sif_df = pd.read_csv(sif_path, sep="\t", header=None, names=["source", "relation", "target"])
    return from_hipathia_dfs(name=name, att_df=att_df, sif_df=sif_df)


def group_delimited_list(entries: List[str], sep: str = "/") -> List[List[str]]:
    """Group delimited things in a list."""
    return [list(b) for a, b in groupby(entries, lambda z: z == sep) if not a]


def _p(identifier: str):
    return Protein(
        namespace="ncbigene",
        identifier=identifier,
        # name=name,
    )


def _f(identifier: str):
    return Protein(
        namespace="hipathia.family",
        identifier=identifier,
        # name=name,
    )


def from_hipathia_dfs(name: str, att_df: pd.DataFrame, sif_df: pd.DataFrame) -> BELGraph:
    """Get a BEL graph from HiPathia dataframes."""

    def _clean_name(s):
        prefix = "N-{name}-".format(name=name)
        if prefix not in s:
            raise ValueError("wrong name for pathway")
        return tuple(sorted(s[len(prefix) :].split(" ")))

    att_df["ID"] = att_df["ID"].map(_clean_name)
    att_df["label"] = att_df["label"].str.split(" ")
    att_df["genesList"] = att_df["genesList"].str.split(",").map(group_delimited_list)

    simple_node_to_dsl = {}
    family_node_to_dsl = {}
    complex_node_to_dsl = {}

    graph = BELGraph(name=name)

    for components, component_label_lists, component_gene_lists in att_df[["ID", "label", "genesList"]].values:
        if not components:
            print(att_df[["ID", "label", "genesList"]])
            raise ValueError("missing components in row")

        if len(components) == 1:  # This is a simple node, representing a protein or protein family
            component, label, entrez_ids = (
                components[0],
                component_label_lists[0],
                component_gene_lists[0],
            )
            if len(entrez_ids) == 1:  # just a protein
                simple_node_to_dsl[component] = _p(identifier=entrez_ids[0])
            else:  # a protein family
                family_dsl = _f(identifier=label)
                for entrez_id in entrez_ids:
                    child_dsl = _p(entrez_id)
                    graph.add_is_a(child_dsl, family_dsl)
                family_node_to_dsl[component] = family_dsl

        else:  # This is a complex node, representing a protein complex of simple nodes
            component_dsls = []
            components = tuple(sorted(components))
            for component, label, entrez_ids in zip(components, component_label_lists, component_gene_lists):
                if len(entrez_ids) == 1:
                    simple_dsl = _p(identifier=entrez_ids[0])
                    simple_node_to_dsl[component] = simple_dsl
                    component_dsls.append(simple_dsl)
                else:
                    family_dsl = _f(identifier=label)
                    for entrez_id in entrez_ids:
                        child_dsl = _p(identifier=entrez_id)
                        graph.add_is_a(child_dsl, family_dsl)
                    family_node_to_dsl[component] = family_dsl
                    component_dsls.append(family_dsl)

            component_dsl = ComplexAbundance(component_dsls)
            graph.add_node_from_data(component_dsl)
            complex_node_to_dsl[components] = component_dsl

    # Remap all of the dictionaries
    x = {}
    x.update(complex_node_to_dsl)
    for k, v in simple_node_to_dsl.items():
        x[(k,)] = v
    for k, v in family_node_to_dsl.items():
        x[(k,)] = v

    sif_df["source"] = sif_df["source"].map(_clean_name).map(x.get)
    sif_df["target"] = sif_df["target"].map(_clean_name).map(x.get)
    for source, relation, target in sif_df.values:
        if relation == "activation":
            graph.add_increases(source, target, citation=(CITATION_TYPE_OTHER, "HiPathia"), evidence="")
        elif relation == "inhibition":
            graph.add_decreases(source, target, citation=(CITATION_TYPE_OTHER, "HiPathia"), evidence="")
        else:
            raise ValueError("unknown relation: {relation}".format(relation=relation))

    return graph


def to_hipathia(
    graph: BELGraph,
    directory: str,
    draw: bool = True,
) -> None:
    """Export HiPathia artifacts for the graph."""
    att_df, sif_df = to_hipathia_dfs(graph, draw_directory=directory if draw else None)
    if att_df is None and sif_df is None:
        logger.warning("can not convert graph %s", graph.name)
        return
    att_df.to_csv(os.path.join(directory, "{}.att".format(graph.name)), sep="\t", index=False)
    sif_df.to_csv(os.path.join(directory, "{}.sif".format(graph.name)), sep="\t", index=False)


def _is_node_family(graph: BELGraph, node: Protein) -> Optional[Set[Protein]]:
    """Get the children of the protein node, if some exist."""
    children = set()
    for child, _, data in graph.in_edges(node, data=True):
        if data[RELATION] == IS_A:
            children.add(child)

    if children and not all(isinstance(child, Protein) for child in children):
        logger.warning("not all children of {} are proteins: {}".format(node, children))
        return

    return children


def to_hipathia_dfs(
    graph: BELGraph,
    draw_directory: Optional[str] = None,
) -> Union[Tuple[None, None], Tuple[pd.DataFrame, pd.DataFrame]]:
    """Get the ATT and SIF dataframes.

    :param graph: A BEL graph
    :param draw_directory: The directory in which a drawing should be output

    1. Identify nodes:
       1. Identify all proteins
       2. Identify all protein families
       3. Identify all complexes with just a protein or a protein family in them
    2. Identify interactions between any of those things that are causal
    3. Profit!
    """
    proteins = set()
    families = defaultdict(set)
    complexes = set()
    for node in sorted(graph, key=str):
        if isinstance(node, Protein):
            children = _is_node_family(graph, node)
            if children:
                families[node] = children
            else:
                proteins.add(node)
        elif isinstance(node, ComplexAbundance) and all(isinstance(m, Protein) for m in node.members):
            complexes.add(node)

    families = {node: sorted(values, key=str) for node, values in sorted(families.items(), key=itemgetter(0))}

    nodes = sorted(proteins.union(families).union(complexes), key=str)
    new_nodes = set()
    edges = []
    for u, v, _, d in sorted(
        graph.out_edges(nodes, keys=True, data=True),
        key=lambda t: (str(t[0]), str(t[1]), t[2]),
    ):
        relation = d[RELATION]
        if relation not in CAUSAL_POLAR_RELATIONS:
            continue
        new_nodes.add(u)
        new_nodes.add(v)
        edges.append(
            (
                u,
                "activation" if relation in CAUSAL_INCREASE_RELATIONS else "inhibition",
                v,
            )
        )

    att = {}
    dsl_to_k = {}
    i = 0
    for node in sorted(new_nodes, key=str):
        if node in families:
            i += 1
            k = (i,)
            children = families[node]
            child_identifiers = [child.identifier for child in children]
            if not all(child_identifiers):
                logger.warning("not all children were grounded: %s", child_identifiers)
                continue
            labels, genes_lists = [node.name], [child_identifiers]
        elif isinstance(node, Protein):
            if not node.identifier or not node.name:
                logger.warning("node was not grounded: %s", node)
                continue
            i += 1
            k = (i,)
            labels, genes_lists = [node.name], [[node.identifier]]
        elif isinstance(node, ComplexAbundance):
            k, labels, genes_lists = [], [], []
            for member in node.members:
                i += 1
                k.append(i)
                labels.append(member.name)
                if member in families:
                    children = families[member]
                    child_identifiers = [child.identifier for child in children]
                    if not all(child_identifiers):
                        logger.warning("not all children were grounded: %s", child_identifiers)
                        continue
                    genes_lists.append(child_identifiers)
                else:
                    if not member.identifier:
                        logger.warning("member was not grounded: %s", member)
                        continue
                    genes_lists.append([member.identifier])
            k = tuple(k)
        else:
            logger.debug("skipping node {}".format(node))
            continue

        k = "N-{}-{}".format(graph.name, " ".join(map(str, k)))
        att[k] = labels, genes_lists
        dsl_to_k[node] = k

    edges = [
        (dsl_to_k[source], relation, dsl_to_k[target])
        for source, relation, target in edges
        if source in dsl_to_k and target in dsl_to_k
    ]
    sif_df = pd.DataFrame(edges)  # DONE

    composite_graph = nx.Graph([(k_source, k_target) for k_source, _, k_target in edges])

    try:
        from networkx.drawing.nx_agraph import pygraphviz_layout

        pos = pygraphviz_layout(composite_graph, prog="neato", args="-Gstart=5")
    except ImportError:
        logger.warning("could not import pygraphviz. Falling back to force directed")
        pos = nx.fruchterman_reingold_layout(composite_graph, seed=5)

    if not pos:
        return None, None

    nx_labels = {}  # from k to label
    min_x = min(x for x, y in pos.values())
    min_y = min(y for x, y in pos.values())

    att_rows = []
    for k, (labels, genes_lists) in sorted(att.items()):
        if k not in pos:
            logger.warning("node not in graph: %s", k)
            continue

        nx_labels[k] = label = " ".join(labels)
        types = ",".join(["gene"] * len(labels))
        gene_list = ",/,".join(",".join(gene_list) for gene_list in genes_lists)
        x, y = pos[k]
        att_rows.append(
            (
                k,  # 1. ID
                label,  # 2. label
                int(100 * (x - min_x)),  # 3. X
                int(100 * (y - min_y)),  # 4. Y
                "white",  # 5. color
                "rectangle",  # 6. shape
                types,  # 7.
                0.5,  # 8. label.cex
                "black",  # 9. label.color
                46,  # 10. width
                17,  # 11. height
                gene_list,  # 12. gene list
            )
        )

    att_df = pd.DataFrame(
        att_rows,
        columns=[
            "ID",
            "label",
            "X",
            "Y",
            "color",
            "shape",
            "type",
            "label.cex",
            "label.color",
            "width",
            "height",
            "genesList",
        ],
    )

    if draw_directory is not None:
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            logger.warning("could not draw graph because matplotlib is not installed")
        else:
            plt.figure(figsize=(20, 20))
            nx.draw_networkx(composite_graph, pos, labels=nx_labels)
            plt.axis("off")
            plt.savefig(os.path.join(draw_directory, "{}.png".format(graph.name)))

    return att_df, sif_df


def make_hsa047370() -> BELGraph:
    """Make an example BEL graph corresponding to the example data from Marina."""
    graph = BELGraph(name="hsa04370")

    node_1 = hgnc(name="CDC42")
    node_9 = hgnc(name="KDR")
    node_11 = hgnc(name="SPHK2")
    node_17 = hgnc(name="MAPKAPK3")
    node_18 = hgnc(name="PPP3CA")
    node_19 = hgnc(name="AKT3")
    node_20 = hgnc(name="PIK3R5")
    node_21 = hgnc(name="NFATC2")
    node_22 = hgnc(name="PRKCA")
    node_24 = hgnc(name="MAPK14")
    node_27 = hgnc(name="SRC")
    node_29 = hgnc(name="VEGFA")
    node_32 = hgnc(name="MAPK1")
    node_33 = hgnc(name="MAP2K1")
    node_34 = hgnc(name="RAF1")
    node_35 = hgnc(name="HRAS")

    node_10 = ComplexAbundance([hgnc(name="PLCG1"), hgnc(name="SH2D2A")])

    node_28 = hgnc(name="SHC2")
    node_23 = hgnc(name="PTK2")
    node_25 = hgnc(name="PXN")
    node_16 = hgnc(name="HSPB1")
    node_36 = hgnc(name="NOS3")
    node_37 = hgnc(name="CASP9")
    node_38 = hgnc(name="BAD")
    node_39 = hgnc(name="RAC1")
    node_14 = hgnc(name="PTGS2")
    node_15 = hgnc(name="PLA2G4B")

    def _add_increases(a, b):
        graph.add_directly_increases(a, b, citation="", evidence="")

    def _add_decreases(a, b):
        graph.add_directly_decreases(a, b, citation="", evidence="")

    _add_increases(node_1, node_24)
    _add_increases(node_9, node_28)

    _add_increases(node_9, node_23)
    _add_increases(node_9, node_25)
    _add_increases(node_9, node_20)
    _add_increases(node_9, node_27)
    _add_increases(node_9, node_10)

    _add_increases(node_11, node_35)
    _add_increases(node_17, node_16)
    _add_increases(node_18, node_21)
    _add_increases(node_19, node_36)
    _add_decreases(node_19, node_37)
    _add_decreases(node_19, node_38)
    _add_increases(node_20, node_39)
    _add_increases(node_20, node_19)
    _add_increases(node_21, node_14)
    _add_increases(node_22, node_34)
    _add_increases(node_22, node_11)
    _add_increases(node_24, node_17)
    _add_increases(node_27, node_20)
    _add_increases(node_29, node_9)
    _add_increases(node_32, node_15)
    _add_increases(node_33, node_32)
    _add_increases(node_34, node_33)
    _add_increases(node_35, node_34)
    _add_increases(node_10, node_18)
    _add_increases(node_10, node_22)
    _add_increases(node_10, node_15)
    _add_increases(node_10, node_36)

    return graph



# -*- coding: utf-8 -*-

"""Test grounding."""

import unittest
from unittest import mock

import bioregistry
import pyobo
from pyobo.mocks import get_mock_id_name_mapping

from pybel.constants import (
    ANNOTATIONS,
    CONCEPT,
    GMOD,
    IDENTIFIER,
    KIND,
    MEMBERS,
    NAME,
    NAMESPACE,
    PMOD,
    VARIANTS,
)
from pybel.grounding import (
    _NAME_REMAPPING,
    _process_annotations,
    _process_concept,
    _process_node,
)
from pybel.language import Entity


def _failer(*_, **__):
    """Fail for all calls to this function."""
    raise ValueError("Called a PyOBO function that should be mocked")


pyobo.getters.get = _failer
pyobo.api.names.cached_mapping = _failer
pyobo.api.names.cached_multidict = _failer

mock_id_name_data = {
    "mesh": {
        "D009474": "Neurons",
        "D010300": "Parkinson Disease",
        "D013378": "Substantia Nigra",
    },
    "doid": {
        "14330": "Parkinson's disease",
    },
    "go": {
        "0006468": "protein phosphorylation",
    },
    "complexportal": {
        "CPX-1829": "Checkpoint clamp complex",
    },
    "ncbitaxon": {
        "9606": "Homo sapiens",
    },
    "cl": {
        "0000030": "glioblast",
    },
    "fplx": {
        "TAP": "TAP",
        "Gamma_secretase": "Gamma_secretase",
    },
}

mock_id_name_mapping = get_mock_id_name_mapping(mock_id_name_data)

_mock_mnemonic_data = {
    "O60921": "HUS1_HUMAN",
    "Q99638": "RAD9A_HUMAN",
    "O60671": "RAD1_HUMAN",
}
_mock_reverse_mnemonic_data = {v: k for k, v in _mock_mnemonic_data.items()}


def _mock_get_mnemonic(identifier, *_, **__):
    return _mock_mnemonic_data[identifier]


mock_get_mnemonic = mock.patch("pybel.grounding.get_mnemonic", side_effect=_mock_get_mnemonic)
mock_get_id_from_mnemonic = mock.patch(
    "pybel.grounding.get_id_from_mnemonic",
    side_effect=_mock_reverse_mnemonic_data.get,
)


@mock_id_name_mapping
@mock_get_mnemonic
@mock_get_id_from_mnemonic
class TestProcessConcept(unittest.TestCase):
    """Test the :func:`_process_concept` function."""

    def _help(self, expected, original, msg=None):
        expected = {CONCEPT: expected}
        d = {CONCEPT: original}
        self.assertIsNotNone(
            bioregistry.normalize_prefix(expected[CONCEPT][NAMESPACE]),
            msg="Unrecognized namespace",
        )
        _process_concept(concept=d[CONCEPT], node=d)
        self.assertEqual(expected[CONCEPT], d[CONCEPT], msg=msg)

    def test_normalize_prefix_case(self, *_):
        """Test normalizing the prefix to the correct case."""
        self._help(
            {NAMESPACE: "mesh", NAME: "Neurons", IDENTIFIER: "D009474"},
            {NAMESPACE: "MESH", NAME: "Neurons", IDENTIFIER: "D009474"},
        )

    def test_normalize_prefix_synonym(self, *_):
        """Test normalizing the prefix based on the synonym dictionary."""
        self._help(
            {NAMESPACE: "mesh", NAME: "Neurons", IDENTIFIER: "D009474"},
            {NAMESPACE: "MESHA", NAME: "Neurons", IDENTIFIER: "D009474"},
        )

    def test_lookup_identifier(self, *_):
        """Test look up of the identifier when given the name."""
        self._help(
            {NAMESPACE: "mesh", NAME: "Neurons", IDENTIFIER: "D009474"},
            {NAMESPACE: "MESH", NAME: "Neurons"},
        )

    def test_lookup_name_as_identifier(self, *_):
        """Test look up of the name when given as the identifier."""
        self._help(
            {NAMESPACE: "mesh", NAME: "Neurons", IDENTIFIER: "D009474"},
            {NAMESPACE: "MESH", IDENTIFIER: "Neurons"},
        )

    def test_lookup_uniprot_identifier(self, *_):
        """Test looking up a uniprot identifier."""
        self._help(
            {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN", IDENTIFIER: "O60921"},
            {NAMESPACE: "UniProt", NAME: "HUS1_HUMAN"},
        )

    def test_fix_uniprot_identifier_as_name(self, *_):
        """Test lookup of the UniProt identifier when given a UniProt identifier as the name."""
        self._help(
            {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN", IDENTIFIER: "O60921"},
            {NAMESPACE: "UniProt", NAME: "O60921"},
        )

    def test_fix_wrong_name(self, *_):
        """Test overwriting a wrong name (not UniProt)."""
        self._help(
            {NAMESPACE: "mesh", NAME: "Neurons", IDENTIFIER: "D009474"},
            {NAMESPACE: "MESH", NAME: "Nonsense name!", IDENTIFIER: "D009474"},
        )

    def test_fix_wrong_uniprot_name(self, *_):
        """Test overwriting a wrong name (UniProt)."""
        self._help(
            {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN", IDENTIFIER: "O60921"},
            {NAMESPACE: "UniProt", NAME: "WRONG!!!!", IDENTIFIER: "O60921"},
        )

    def test_remap_sfam(self, *_):
        """Test remapping SFAM to FPLX."""
        self.assertIn(("bel", "TAP Family"), _NAME_REMAPPING)
        self._help(
            {NAMESPACE: "fplx", NAME: "TAP", IDENTIFIER: "TAP"},
            {NAMESPACE: "SFAM", NAME: "TAP Family"},
        )

    def test_remap_scomp(self, *_):
        """Test remapping SFAM to FPLX."""
        self.assertIsNotNone(bioregistry.normalize_prefix("BEL"))
        self.assertIn(
            ("bel", "gamma Secretase Complex"),
            _NAME_REMAPPING,
            msg="name remapping is not populated properly",
        )
        self._help(
            {NAMESPACE: "fplx", NAME: "Gamma_secretase", IDENTIFIER: "Gamma_secretase"},
            {NAMESPACE: "SCOMP", NAME: "gamma Secretase Complex"},
        )


@mock_id_name_mapping
@mock_get_mnemonic
@mock_get_id_from_mnemonic
class TestGround(unittest.TestCase):
    """Test grounding."""

    def _help(self, expected, result):
        _process_node(result)
        self.assertEqual(expected, result)

    def test_lookup_identifier_member(self, *_):
        """Test looking up the identifier of a member by name."""
        self._help(
            {
                MEMBERS: [
                    {
                        CONCEPT: {
                            NAMESPACE: "mesh",
                            NAME: "Neurons",
                            IDENTIFIER: "D009474",
                        }
                    }
                ]
            },
            {MEMBERS: [{CONCEPT: {NAMESPACE: "MESH", NAME: "Neurons"}}]},
        )

    def test_lookup_identifier_complex(self, *_):
        """Test looking up the identifier of a named complex and its members at the same time."""
        self._help(
            {
                CONCEPT: {
                    NAMESPACE: "complexportal",
                    NAME: "Checkpoint clamp complex",
                    IDENTIFIER: "CPX-1829",
                },
                MEMBERS: [
                    {
                        CONCEPT: {
                            NAMESPACE: "uniprot",
                            NAME: "HUS1_HUMAN",
                            IDENTIFIER: "O60921",
                        }
                    },
                    {
                        CONCEPT: {
                            NAMESPACE: "uniprot",
                            NAME: "RAD9A_HUMAN",
                            IDENTIFIER: "Q99638",
                        }
                    },
                    {
                        CONCEPT: {
                            NAMESPACE: "uniprot",
                            NAME: "RAD1_HUMAN",
                            IDENTIFIER: "O60671",
                        }
                    },
                ],
            },
            {
                CONCEPT: {NAMESPACE: "complexportal", NAME: "Checkpoint clamp complex"},
                MEMBERS: [
                    {CONCEPT: {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN"}},
                    {CONCEPT: {NAMESPACE: "uniprot", NAME: "RAD9A_HUMAN"}},
                    {CONCEPT: {NAMESPACE: "uniprot", NAME: "RAD1_HUMAN"}},
                ],
            },
        )

    def test_lookup_identifier_protein(self, *_):
        """Test looking up the identifier based on a protein's name."""
        self._help(
            {CONCEPT: {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN", IDENTIFIER: "O60921"}},
            {CONCEPT: {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN"}},
        )

    def test_lookup_name_protein(self, *_):
        """Test looking up the name based on a protein's identifier."""
        self._help(
            {CONCEPT: {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN", IDENTIFIER: "O60921"}},
            {CONCEPT: {NAMESPACE: "uniprot", IDENTIFIER: "O60921"}},
        )

    def test_fix_name_protein(self, *_):
        """Test fixing a wrong name by overwriting by identifier-based lookup."""
        self._help(
            {CONCEPT: {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN", IDENTIFIER: "O60921"}},
            {CONCEPT: {NAMESPACE: "uniprot", IDENTIFIER: "O60921", NAME: "wrong!!!"}},
        )

    def test_lookup_identifier_pmod(self, *_):
        """Test looking up a protein modification's identifier by name."""
        self._help(
            {
                CONCEPT: {
                    NAMESPACE: "uniprot",
                    NAME: "HUS1_HUMAN",
                    IDENTIFIER: "O60921",
                },
                VARIANTS: [
                    {
                        KIND: PMOD,
                        CONCEPT: {
                            NAMESPACE: "go",
                            IDENTIFIER: "0006468",
                            NAME: "protein phosphorylation",
                        },
                    },
                ],
            },
            {
                CONCEPT: {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN"},
                VARIANTS: [
                    {
                        KIND: PMOD,
                        CONCEPT: {NAMESPACE: "GO", NAME: "protein phosphorylation"},
                    },
                ],
            },
        )

    def test_lookup_name_pmod(self, *_):
        """Test looking up a protein modification's name by identifier."""
        self._help(
            {
                CONCEPT: {
                    NAMESPACE: "uniprot",
                    NAME: "HUS1_HUMAN",
                    IDENTIFIER: "O60921",
                },
                VARIANTS: [
                    {
                        KIND: PMOD,
                        CONCEPT: {
                            NAMESPACE: "go",
                            IDENTIFIER: "0006468",
                            NAME: "protein phosphorylation",
                        },
                    },
                ],
            },
            {
                CONCEPT: {NAMESPACE: "uniprot", IDENTIFIER: "O60921"},
                VARIANTS: [
                    {
                        KIND: PMOD,
                        CONCEPT: {NAMESPACE: "GO", IDENTIFIER: "0006468"},
                    },
                ],
            },
        )

    def test_fix_pmod_name(self, *_):
        """Test fixing a wrong name in a pmod."""
        self._help(
            {
                CONCEPT: {
                    NAMESPACE: "uniprot",
                    NAME: "HUS1_HUMAN",
                    IDENTIFIER: "O60921",
                },
                VARIANTS: [
                    {
                        KIND: PMOD,
                        CONCEPT: {
                            NAMESPACE: "go",
                            IDENTIFIER: "0006468",
                            NAME: "protein phosphorylation",
                        },
                    },
                ],
            },
            {
                CONCEPT: {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN"},
                VARIANTS: [
                    {
                        KIND: PMOD,
                        CONCEPT: {
                            NAMESPACE: "GO",
                            IDENTIFIER: "0006468",
                            NAME: "WRONG!",
                        },
                    },
                ],
            },
        )

    def test_normalize_pmod_default(self, *_):
        """Test normalizing a pmod using the default bel namespace."""
        self._help(
            {
                CONCEPT: {
                    NAMESPACE: "uniprot",
                    NAME: "HUS1_HUMAN",
                    IDENTIFIER: "O60921",
                },
                VARIANTS: [
                    {
                        KIND: PMOD,
                        CONCEPT: {
                            NAMESPACE: "go",
                            IDENTIFIER: "0006468",
                            NAME: "protein phosphorylation",
                        },
                    },
                ],
            },
            {
                CONCEPT: {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN"},
                VARIANTS: [
                    {
                        KIND: PMOD,
                        CONCEPT: {NAMESPACE: "bel", NAME: "Ph"},
                    },
                ],
            },
        )

    def test_normalize_pmod_default_methylation(self, *_):
        """Test normalizing the default namespace's Me entry because of conflict with gmods."""
        self._help(
            {
                CONCEPT: {
                    NAMESPACE: "uniprot",
                    NAME: "HUS1_HUMAN",
                    IDENTIFIER: "O60921",
                },
                VARIANTS: [
                    {
                        KIND: PMOD,
                        CONCEPT: {
                            NAMESPACE: "go",
                            IDENTIFIER: "0006479",
                            NAME: "protein methylation",
                        },
                    },
                ],
            },
            {
                CONCEPT: {NAMESPACE: "uniprot", NAME: "HUS1_HUMAN"},
                VARIANTS: [
                    {KIND: PMOD, CONCEPT: {NAMESPACE: "bel", NAME: "Me"}},
                ],
            },
        )

    def normalize_gmod_default_methylation(self, *_):
        """Test normalizing the default namespace's Me entry because of conflict with pmods."""
        self._help(
            {
                CONCEPT: {NAMESPACE: "hgnc", NAME: "MAPT", IDENTIFIER: "6893"},
                VARIANTS: [
                    {
                        CONCEPT: {
                            NAMESPACE: "go",
                            IDENTIFIER: "0006306",
                            NAME: "DNA methylation",
                        },
                        KIND: GMOD,
                    },
                ],
            },
            {
                CONCEPT: {NAMESPACE: "HGNC", NAME: "MAPT"},
                VARIANTS: [
                    {CONCEPT: {NAMESPACE: "bel", NAME: "Me"}, KIND: GMOD},
                ],
            },
        )


@mock_id_name_mapping
class TestAnnotations(unittest.TestCase):
    """Test processing annotations."""

    def _help(self, expected_data, data):
        expected_data = {ANNOTATIONS: expected_data}
        data = {ANNOTATIONS: data}
        _process_annotations(data)
        self.assertEqual(expected_data, data)

    def test_lookup_by_identifier(self, *_):
        """Test lookup by identifier."""
        self._help(
            {"Disease": [Entity(namespace="mesh", identifier="D010300", name="Parkinson Disease")]},
            {"Disease": [Entity(namespace="mesh", identifier="D010300")]},
        )

    def test_lookup_by_name(self, *_):
        """Test lookup by name."""
        self._help(
            {"Disease": [Entity(namespace="mesh", identifier="D010300", name="Parkinson Disease")]},
            {"Disease": [Entity(namespace="mesh", name="Parkinson Disease")]},
        )

    def test_lookup_by_name_as_identifier(self, *_):
        """Test lookup by name if it's accidentally in the identifier slot."""
        self._help(
            {"Disease": [Entity(namespace="mesh", identifier="D010300", name="Parkinson Disease")]},
            {"Disease": [Entity(namespace="mesh", identifier="Parkinson Disease")]},
        )

    def test_upgrade_category(self, *_):
        """Test upgrading the category."""
        self._help(
            {"Disease": [Entity(namespace="mesh", identifier="D010300", name="Parkinson Disease")]},
            {"MeSHDisease": [Entity(namespace="mesh", identifier="D010300", name="Parkinson Disease")]},
        )

    def test_upgrade_category_and_namespace(self, *_):
        """Test upgrading the category and the namespace simultaneously."""
        self._help(
            {"Disease": [Entity(namespace="mesh", identifier="D010300", name="Parkinson Disease")]},
            {
                "MeSHDisease": [
                    Entity(
                        namespace="MeSHDisease",
                        identifier="D010300",
                        name="Parkinson Disease",
                    )
                ]
            },
        )

    def test_upgrade_with_name_as_identifier(self, *_):
        """Test upgrading MeSH disease, MeSH anatomy, and Species tags and lookup by name, in the identifiers space."""
        self._help(
            {  # Expected
                "Disease": [Entity(namespace="mesh", identifier="D010300", name="Parkinson Disease")],
                "Anatomy": [Entity(namespace="mesh", identifier="D013378", name="Substantia Nigra")],
                "Species": [Entity(namespace="ncbitaxon", identifier="9606", name="Homo sapiens")],
            },
            {  # Original
                "MeSHDisease": [Entity(namespace="MeSHDisease", identifier="Parkinson Disease")],
                "MeSHAnatomy": [Entity(namespace="MeSHAnatomy", identifier="Substantia Nigra")],
                "Species": [Entity(namespace="Species", identifier="Homo sapiens")],
            },
        )

    def test_upgrade_by_identifier(self, *_):
        """Test upgrading and lookup by identifier."""
        self._help(
            {  # Expected
                "Disease": [Entity(namespace="mesh", identifier="D010300", name="Parkinson Disease")],
                "Anatomy": [Entity(namespace="mesh", identifier="D013378", name="Substantia Nigra")],
                "Species": [Entity(namespace="ncbitaxon", identifier="9606", name="Homo sapiens")],
            },
            {  # Original
                "MeSHDisease": [Entity(namespace="MeSHDisease", identifier="D010300")],
                "MeSHAnatomy": [Entity(namespace="MeSHAnatomy", identifier="D013378")],
                "Species": [Entity(namespace="Species", identifier="9606")],
            },
        )

    def test_upgrade_by_name(self, *_):
        """Test upgrading and lookup by name."""
        self._help(
            {  # Expected
                "Disease": [Entity(namespace="mesh", identifier="D010300", name="Parkinson Disease")],
                "Anatomy": [Entity(namespace="mesh", identifier="D013378", name="Substantia Nigra")],
                "Species": [Entity(namespace="ncbitaxon", identifier="9606", name="Homo sapiens")],
            },
            {  # Original
                "MeSHDisease": [Entity(namespace="MeSHDisease", name="Parkinson Disease")],
                "MeSHAnatomy": [Entity(namespace="MeSHAnatomy", name="Substantia Nigra")],
                "Species": [Entity(namespace="Species", name="Homo sapiens")],
            },
        )

    def test_unmappable_category(self, *_):
        """Test when the category can't be mapped."""
        self._help(
            {  # Expected
                "Custom Annotation": [Entity(namespace="Custom Annotation", identifier="Custom Value")],
            },
            {
                "Custom Annotation": [Entity(namespace="Custom Annotation", identifier="Custom Value")],
            },
        )

    def test_unmappable_identifier(self, *_):
        """Test when the identifier can not be resolved."""
        self._help(
            {  # Expected
                "Disease": [Entity(namespace="doid", identifier="Failure")],
            },
            {
                "Disease": [Entity(namespace="Disease", identifier="Failure")],
            },
        )

    def test_unmappable_name(self, *_):
        """Test when the identifier can not be looked up by name."""
        self._help(
            {  # Expected
                "Disease": [Entity(namespace="doid", name="Failure")],
            },
            {
                "Disease": [Entity(namespace="Disease", name="Failure")],
            },
        )



# -*- coding: utf-8 -*-

"""Tests for the mocks for the query builder."""

import unittest

from pybel.examples import egf_graph
from pybel.testing.mock_manager import MockQueryManager


class TestMockManager(unittest.TestCase):
    """Tests for the mock query manager."""

    def test_make(self):
        """Test instantiating the mock query manager."""
        manager = MockQueryManager()
        self.assertEqual(0, manager.count_networks())

    def test_make_with_graph(self):
        """Test counting networks in the mock query manager."""
        manager = MockQueryManager(graphs=[egf_graph])
        self.assertEqual(1, manager.count_networks())

    def test_add_graph(self):
        """Test adding a graph with insert_graph."""
        manager = MockQueryManager()
        graph = egf_graph.copy()
        manager.insert_graph(graph)
        self.assertEqual(1, manager.count_networks())
